# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

"""
Parsers for MTP/Mlip related files.
"""

import pyparsing as pp
import numpy as np

__author__ = "Marvin Poul"
__copyright__ = "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - " \
                "Computational Materials Design (CM) Department"
__version__ = "1.0"
__maintainer__ = "Marvin Poul"
__email__ = "poul@mpie.de"
__status__ = "development"
__date__ = "Aug 18, 2021"

pp.ParserElement.setDefaultWhitespaceChars(" ")

def _make_potential_parser():
    NL = pp.Suppress("\n")
    EQ = pp.Suppress("=")
    LB = pp.Suppress("{")
    RB = pp.Suppress("}")

    def make_keyword(name):
        return pp.Suppress(pp.Keyword(name))

    def make_field(name, expr, key=None, ungroup=True):
        key = name if key is None else key
        field_expr = make_keyword(name) + EQ + expr + NL[0,1]
        if ungroup:
            field_expr = pp.ungroup(field_expr)
        field_expr = field_expr.set_results_name(key)
        field_expr.set_name(key)
        return field_expr

    def make_list(field_expr, grouped=False):
        list_expr = LB + pp.delimited_list(field_expr) + RB
        if grouped:
            list_expr = pp.Group(list_expr)
        list_expr = list_expr.set_name("list")
        return list_expr

    radial_basis_type = make_field("radial_basis_type", "RBChebyshev", "basis_type")
    radial_info = pp.IndentedBlock(
            make_field("min_dist", pp.pyparsing_common.fnumber) \
          + make_field("max_dist", pp.pyparsing_common.fnumber) \
          + make_field("radial_basis_size", pp.pyparsing_common.integer, "basis_size") \
          + make_field("radial_funcs_count", pp.pyparsing_common.integer, "funcs_count"),
    ).set_results_name("info")
    radial_info.set_parse_action(lambda tk: tk[0].as_dict())

    radial_func_types = pp.Word(pp.nums) + pp.Suppress("-") + pp.Word(pp.nums) + NL
    radial_func_types.set_name("radial function types")
    radial_func_types.set_parse_action(lambda tokens: f"{tokens[0]}-{tokens[1]}")
    radial_func_coeffs = make_list(pp.pyparsing_common.fnumber, grouped=True)
    radial_funcs = pp.IndentedBlock(
            make_keyword("radial_coeffs") + NL \
            + pp.IndentedBlock(
                    radial_func_types \
                            + pp.IndentedBlock(radial_func_coeffs + NL)[1, ...]
              )[1, ...]
            )

    radial_funcs = radial_funcs.set_results_name("funcs")
    radial_funcs = pp.ungroup(radial_funcs).set_results_name("funcs")
    radial_funcs.set_parse_action(lambda tokens: {k: v for k, v in tokens.as_list()[0]})

    radial = pp.Group(
              radial_basis_type \
            + radial_info \
            + radial_funcs)
    radial = radial.set_results_name("radial")

    MTP = make_keyword("MTP") + NL

    parser = NL[0, ...] + MTP + pp.Each([
        make_field( "version",        pp.Word(pp.nums + ".") ),
        make_field( "potential_name", pp.Word(pp.alphanums) ),
        make_field( "scaling",        pp.pyparsing_common.fnumber ),
        make_field( "species_count",  pp.pyparsing_common.integer ),
        make_field( "potential_tag",  pp.Optional(pp.Word(pp.alphanums), "")  ),
        radial,
        make_field( "alpha_moments_count", pp.pyparsing_common.integer ),
        make_field( "alpha_index_basic_count", pp.pyparsing_common.integer ),
        make_field( "alpha_index_basic", make_list(make_list(pp.pyparsing_common.integer, grouped=True)), ungroup=False ),
        make_field( "alpha_index_times_count", pp.pyparsing_common.integer ),
        make_field( "alpha_index_times", make_list(make_list(pp.pyparsing_common.integer, grouped=True)), ungroup=False ),
        make_field( "alpha_scalar_moments", pp.pyparsing_common.integer ),
        make_field( "alpha_moment_mapping", make_list(pp.pyparsing_common.integer), ungroup=False),
        make_field( "species_coeffs", make_list(pp.pyparsing_common.fnumber), ungroup=False ),
        make_field( "moment_coeffs", make_list(pp.pyparsing_common.fnumber), ungroup=False ),
    ])

    return parser

def potential(potential_string):
    """
    Parse an MTP potential for mlip.

    Args:
        potential_string (str): string to parse

    Raises:
        ValueError: failed to parse potential
    """
    try:
        result = _make_potential_parser().parse_string(potential_string).as_dict()
        result["radial"]["basis_type"] = result["radial"]["basis_type"][2:] # strip RB prefix
        # Convert to numpy arrays
        for pair, func in result["radial"]["funcs"].items():
            result["radial"]["funcs"][pair] = np.array(result["radial"]["funcs"][pair])
        result["alpha_index_basic"] = np.array(result["alpha_index_basic"])
        result["alpha_index_times"] = np.array(result["alpha_index_times"])
        result["alpha_moment_mapping"] = np.array(result["alpha_moment_mapping"])
        result["species_coeffs"] = np.array(result["species_coeffs"])
        result["moment_coeffs"] = np.array(result["moment_coeffs"])
        basis_size = result["radial"]["info"]["basis_size"]
        funcs_size = result["radial"]["info"]["funcs_count"]
        for pair, func in result["radial"]["funcs"].items():
            if func.shape != (funcs_size, basis_size):
                raise ValueError(f"Invalid radial basis for pair {pair}, should be {funcs_count}x{ basis_size} not {result['radial']['funcs'].shape}")
        if result['alpha_index_basic'].shape != (result['alpha_index_basic_count'], 4):
            raise ValueError(f"Invalid alpha basic indices, length should be {result['alpha_index_basic_count']}")
        if result["alpha_index_times"].shape != (result["alpha_index_times_count"], 4):
            raise ValueError(f"Invalid alpha times indices, length should be {result['alpha_index_times_count']}")
        if len(result["alpha_moment_mapping"]) != result["alpha_scalar_moments"]:
            raise ValueError(f"Invalid alpha moment mapping, length should be {result['alpha_scalar_moments']}")
        if len(result["moment_coeffs"]) != result["alpha_scalar_moments"]:
            raise ValueError(f"Invalid moment coefficients, length should be {result['alpha_scalar_moments']}")
        if len(result["species_coeffs"]) != result["species_count"]:
            raise ValueError(f"Invalid species coefficients, length should be {result['species_count']}")

        return result
    except pp.ParseException:
        raise ValueError("failed to parse potential") from None



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import numpy as np
import random
import scipy.constants as sci_const
from pyiron_base import GenericParameters
from pyiron.lammps.interactive import LammpsInteractive
from pyiron.atomistics.job.interactivewrapper import InteractiveWrapper
from pyiron.atomistics.structure.atoms import Atoms

__author__ = "Jan Janssen"
__copyright__ = "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - " \
                "Computational Materials Design (CM) Department"
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "development"
__date__ = "Jul 25, 2018"


class MonteCarloMaster(InteractiveWrapper):
    """
    A reference implementation for Lattice Monte Carlo Simulation. To understand Monte Carlo Simulation we highly
    recommend implementing your own Monte Carlo Method, still this is a reference implementation to compare your
    results. On a predefined lattice, the atoms are inserted as interstitials and the energy is calculated using
    a quantum engine, which can be set as reference job.

    Args:
        project (ProjectHDFio): ProjectHDFio instance which points to the HDF5 file the job is stored in
        job_name (str): name of the job, which has to be unique within the project

    Attributes:
        input (pyiron_base.GenericParameters): handles the input
        structure_lattice (pyiron.atomistics.structure.atoms.Atoms): Crystal Lattice
        structure_sublattice (pyiron.atomistics.structure.atoms.Atoms): Crystal Lattice of interstitial positions
        ref_job (pyiron.atomistics.job.atomistic.AtomisticGenericJob): atomistic interpreter used to evaluate structures
    """
    def __init__(self, project, job_name):
        super(MonteCarloMaster, self).__init__(project, job_name)
        self.__name__ = "MonteCarloMaster"
        self.input = MonteCarloInput()
        self._structure_sublattice = None
        self._structure_lattice = None

    @property
    def structure_sublattice(self):
        """
        Atomistic crystal structure of all possible interstital positions
        """
        return self._structure_sublattice

    @structure_sublattice.setter
    def structure_sublattice(self, basis):
        """
        Atomistic crystal structure of all possible interstital positions

        Args:
            basis (pyiron.atomistics.structure.atoms.Atoms): Atomistic crystal structure
        """
        self._structure_sublattice = basis

    @property
    def structure_lattice(self):
        """
        Lattice structure - defining the matrix
        """
        return self._structure_lattice

    @structure_lattice.setter
    def structure_lattice(self, basis):
        """
        Lattice structure - defining the matrix

        Args:
            basis (pyiron.atomistics.structure.atoms.Atoms): Atomistic crystal structure
        """
        self._structure_lattice = basis

    def set_input_to_read_only(self):
        """
        This function enforces read-only mode for the input classes, but it has to be implement in the individual
        classes.
        """
        super(MonteCarloMaster, self).set_input_to_read_only()
        self.input.read_only = True

    def validate_ready_to_run(self):
        """
        Validate the calculation is ready to run - in particular check that both the lattice structure and the
        sublattice structure are defined.
        """
        super(MonteCarloMaster, self).validate_ready_to_run()
        if not self.structure_sublattice:
            raise ValueError('This job does not contain a valid intersitial structure: {}'.format(self.job_name))

    def db_entry(self):
        """
        Generate the database entry for the monte carlo calculation, based on the initial concentration

        Returns:
            (dict): db_dict
        """
        db_dict = super(MonteCarloMaster, self).db_entry()
        if self.structure_lattice and self.structure_sublattice:
            interstitial_selected_lst = self._initial_config(
                self.input['number_of_interstitials'],
                self.structure_sublattice
            )
            structure = self.structure_lattice + self.structure_sublattice[interstitial_selected_lst]
            parent_structure = structure.get_parent_basis()
            db_dict["ChemicalFormula"] = parent_structure.get_chemical_formula()
        return db_dict

    def to_hdf(self, hdf=None, group_name=None):
        """
        Store the ExampleJob object in the HDF5 File

        Args:
            hdf (ProjectHDFio): HDF5 group object - optional
            group_name (str): HDF5 subgroup name - optional
        """
        super(MonteCarloMaster, self).to_hdf(hdf=hdf, group_name=group_name)
        with self.project_hdf5.open("input") as hdf5_input:
            self.structure_sublattice.to_hdf(hdf5_input, group_name="structure_sublattice")
            self.structure_lattice.to_hdf(hdf5_input, group_name="structure_lattice")

    def from_hdf(self, hdf=None, group_name=None):
        """
        Restore the ExampleJob object in the HDF5 File

        Args:
            hdf (ProjectHDFio): HDF5 group object - optional
            group_name (str): HDF5 subgroup name - optional
        """
        super(MonteCarloMaster, self).from_hdf(hdf=hdf, group_name=group_name)
        with self.project_hdf5.open("input") as hdf5_input:
            self.structure_sublattice = Atoms().from_hdf(hdf5_input, group_name="structure_sublattice")
            self.structure_lattice = Atoms().from_hdf(hdf5_input, group_name="structure_lattice")

    def write_input(self):
        """
        No input files are written - they are not required for this job
        """
        pass

    def _write_run_wrapper(self, debug=False):
        """
        No wrapper is written - it is not required for this job

        Args:
            debug (bool): the debug flag is ignored.
        """
        pass

    def run_static(self):
        """
        Loop over the temperatures defined in the input and for each temperature calculate n rounds as defined in the
        input, finally during each round call the monte carlo calculation.
        """
        for temperature in self.input['temperatures']:
            for round_nr in range(int(self.input['run_rounds'])):
                self.status.running = True
                interstitial_selected_lst = self._initial_config(
                    self.input['number_of_interstitials'],
                    self.structure_sublattice
                )
                energy_accepted_lst, energy_total_lst = self._calc_monte_carlo(
                    interstitial_selected_lst,
                    temperature
                )
                if self.ref_job.server.run_mode.interactive:
                    self.ref_job.interactive_close()
                self.status.collect = True
                self._store_output_in_hdf5(temperature, round_nr, energy_accepted_lst, energy_total_lst)
        self._finish_job()

    def _calc_monte_carlo(self, interstitial_selected_lst, temperature):
        """
        Internal helper function to execute one monte carlo run, loop over the number of steps and at each step,
        test a configuration, by randomly choosing a subset of the available indices and populating those with
        interstitials. The execution is finished once either the number of accepted steps equals the run_steps
        defined in the input or once the number of total calculation equals the product of run_steps and
        run_try_factor.

        Args:
            interstitial_selected_lst (list): possible interstitial indices
            temperature (float): temperature which defines the acceptance ratio

        Returns:
            (list): accepted energies
            (list): all energies
        """
        step_nr, try_nr = 0, 0
        e_prev = None
        energy_total_lst, energy_accepted_lst = [], []
        while step_nr < int(self.input['run_steps']) and \
                try_nr < int(self.input['run_steps']) * int(self.input['run_try_factor']):
            prev_positions = interstitial_selected_lst[:]

            interstitial_selected_lst = self._index_switch(interstitial_selected_lst)
            structure = self.structure_lattice + self.structure_sublattice[interstitial_selected_lst]
            e_step = self._calc_energy_for_structure(structure)

            if self._monte_carlo_accept_step(temperature, e_step, e_prev):
                e_prev = e_step
                energy_accepted_lst.append(e_step)
                step_nr += 1
            else:
                interstitial_selected_lst = prev_positions[:]
            energy_total_lst.append(e_step)
            try_nr += 1
        return energy_accepted_lst, energy_total_lst

    def _index_switch(self, interstitial_selected_lst):
        """
        Internal helper function to select a subset of the available interstitial positions

        Args:
            interstitial_selected_lst (list): index of possible positions

        Returns:
            list: random selection from the possible possitions
        """
        interstitial_not_selected_lst = self._not_selected(
            interstitial_selected_lst=interstitial_selected_lst,
            sub_lattice_structure=self.structure_sublattice
        )
        random_new_element = random.choice(interstitial_not_selected_lst)
        random_current_element = random.choice(interstitial_selected_lst)
        del interstitial_selected_lst[interstitial_selected_lst.index(random_current_element)]
        interstitial_selected_lst.append(random_new_element)
        return interstitial_selected_lst

    def _calc_energy_for_structure(self, structure):
        """
        Calculate the energy using the quantum engine defined as ref_job

        Args:
            structure (pyiron.atomistics.structure.atoms.Atoms): Atomistic crystal structure to evaluate

        Returns:
            float: energy calculated
        """
        if isinstance(self.ref_job, LammpsInteractive) and len(self._job_name_lst) == 0 and self.ref_job.server.run_mode.interactive:
            self.ref_job.interactive_positions_setter(structure.positions)
            self.ref_job._interactive_lib_command(self.ref_job._interactive_run_command)
            return self.ref_job.interactive_energy_tot_getter()
        else:
            self.ref_job.structure = structure
            self.ref_job_initialize()
            if self.ref_job.server.run_mode.interactive:
                self.ref_job.run()
            else:
                self.ref_job.run(run_again=True)
            return self.ref_job.output.energy_tot[-1]

    def _store_output_in_hdf5(self, temperature, round_nr, energy_accepted_lst, energy_total_lst):
        """
        Store accepted energies and total energies inside the HDF5 file

        Args:
            temperatur (int/float): current temperature
            round_nr (int): current round
            energy_accepted_lst (list): list of accepted energies
            energy_total_lst (list): list of total energies
        """
        with self.project_hdf5.open("output") as h5:
            h5['energy_accepted_' + str(int(temperature)) + '_' + str(int(round_nr))] = \
                np.array(energy_accepted_lst)
            h5['energy_total_' + str(int(temperature)) + '_' + str(int(round_nr))] = np.array(energy_total_lst)

    @staticmethod
    def _monte_carlo_accept_step(temperature, e_step, e_prev):
        """
        Monte carlo step - decide whether a new step is accepted or not.

        Args:
            temperature (temperature): Temperature of the Monte Carlo Calculation
            e_step (flaot): Energy of the current step
            e_prev (float): Energy of the previous step

        Returns:
            boolean
        """
        if e_prev:
            energy_change = e_step - e_prev
            exponent = -energy_change / (sci_const.Boltzmann / sci_const.e * temperature)
            if exponent > 0:  # np.exp(709) = 8.2184074615549724e+307
                probability = 1.0  # np.exp(710) = inf  ('RuntimeWarning: overflow encountered in exp')
            else:
                probability = np.exp(exponent)
            return random.random() <= probability or energy_change < 0
        else:
            return True

    @staticmethod
    def _not_selected(interstitial_selected_lst, sub_lattice_structure):
        """
        Internal helper function to identify the interstitial opositions which are not selected

        Args:
            interstitial_selected_lst (list): selected positions
            sub_lattice_structure (pyiron.atomistics.structure.atoms.Atoms): atomic crystal structure of all
                possible interstitial positions

        Returns:
            list: positions which are not selected
        """
        interstitial_all_lst = list(range(len(sub_lattice_structure)))
        return list(set(interstitial_all_lst) - set(interstitial_selected_lst))

    @staticmethod
    def _initial_config(number_of_interstitials, sublattice):
        """
        Internal helper function to generate the initial configuration

        Args:
            number_of_interstitials (int): number of interstitials
            sublattice (pyiron.atomistics.structure.atoms.Atoms): atomic crystal structure of all possible
                interstitial positions

        Returns:
            list: random index of interstitial positions
        """
        interstitial_index = []
        while len(interstitial_index) < number_of_interstitials:
            interstitial_index.append(random.randint(a=0, b=len(sublattice) - 1))
            interstitial_index = list(set(interstitial_index))
        return interstitial_index


class MonteCarloInput(GenericParameters):
    """
    Input class for the ExampleJob based on the GenericParameters class.

    Args:
        input_file_name (str): Name of the input file - optional
    """
    def __init__(self, input_file_name=None):
        super(MonteCarloInput, self).__init__(
            input_file_name=input_file_name,
            table_name="montecarlo_inp",
            separator_char="=",
            comment_char="#"
        )

    def load_default(self):
        """
        Loading the default settings for the input file.
        """
        input_str = '''\
number_of_interstitials = 32
run_rounds = 1
run_steps = 100
run_try_factor = 10
temperatures = [200]
'''
        self.load_string(input_str)



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
import datetime
from getpass import getpass

from dateutil import parser
import os.path
import io
import warnings

import pyiron_base
from pyiron_base import state, ProjectHDFio
from pyiron_base.interfaces.has_groups import HasGroups
from pyiron_contrib.generic.filedata import StorageInterface
from pyiron_contrib.generic.filedata import (
    FileDataTemplate,
    load_file,
    MetaDataTemplate,
)
from typing import Union, List

import coscine


class CoscineMetadata(coscine.resource.MetadataForm, MetaDataTemplate):
    """Add a proper representation to the coscine version"""

    def __init__(self, meta_data_form: coscine.resource.MetadataForm):
        super().__init__(client=meta_data_form.client, graph=meta_data_form.profile)

    def to_dict(self):
        result = {}
        for key, value in self.items():
            if len(str(value)) > 0 and not self.is_controlled(key):
                result[key] = value.raw()
            elif len(str(value)) > 0:
                result[key] = str(value)
        return result

    def __repr__(self):
        return self.__str__()

    def __setitem__(self, key, value):
        try:
            super().__setitem__(key, value)
        except TypeError as e:
            val = parser.parse(value)
            try:
                super().__setitem__(key, val)
            except TypeError:
                raise e


class CoscineFileData(FileDataTemplate):
    def __init__(self, coscine_object: coscine.FileObject):
        super().__init__()
        self._coscine_object = coscine_object
        self._data = None
        self._filename = coscine_object.name
        self.filetype = self._get_filetype_from_filename(self._filename)

    @property
    def data(self):
        if self._data is None:
            self.load_data()
            # return "Data not yet loaded! Call `load_data` to load"
        if isinstance(self._data, str):
            return load_file(self._data, filetype=self.filetype)
        return load_file(io.BytesIO(self._data), filetype=self.filetype)

    def load_data(self, force_update=False):
        if self._data is None or force_update:
            if (
                "Software IDs" in self.metadata
                and self.metadata["Software IDs"].raw().lower().startswith("pyiron")
                and self.filetype in ["h5", "hdf"]
            ):
                tmp_dir = os.path.join(os.curdir, "coscine_downloaded_h5")
                if not os.path.exists(tmp_dir):
                    os.mkdir(tmp_dir)
                self.download(tmp_dir)
                file_name = os.path.abspath(os.path.join(tmp_dir, self._filename))
                if len(self.metadata["External/alias ID"].raw()) > 0:
                    new_name = os.path.abspath(
                        os.path.join(
                            tmp_dir,
                            "pyiron_"
                            + self.metadata["External/alias ID"].raw()
                            + ".h5",
                        )
                    )
                    os.rename(file_name, new_name)
                    file_name = new_name
                self._data = file_name
            else:
                self._data = self._coscine_object.content()

    def download(self, path="./"):
        self._coscine_object.download(path=path)

    def _get_metadata(self):
        form = CoscineMetadata(self._coscine_object.form())
        form.parse(self._coscine_object.metadata(force_update=True))
        return form

    def _set_metadata(self, metadata):
        raise AttributeError("can't set attribute")


def _list_filter(list_to_filter, **kwargs):
    """Return entry in list matching all attributes."""

    def filter_func(item):
        for key, value in kwargs.items():
            if getattr(item, key) != value:
                return False
        return True

    return list(filter(filter_func, list_to_filter))


class Job2CoscineMetadataConverter:

    _known_profiles = [
        "https://purl.org/coscine/ap/sfb1394/AtomisticSimulation/",
        "https://purl.org/coscine/ap/sfb1394/AtomisticSimulationMD/",
    ]

    def __init__(self, job, form):
        form_profile = form.profile.target()
        if "sfb1394" not in form_profile:
            raise ValueError("Unknown form, not a SFB1394 one.")

        self._job = job
        self._form = form

        user = job.project_hdf5["server"]["user"]
        now = datetime.datetime.today()

        self._form["ID"] = job.name
        self._form["External/alias ID"] = str(job.id)
        self._form["User"] = user
        self._form["Date"] = now

        if form_profile in self._known_profiles:
            self._form["Software IDs"] = "pyiron"
            self._parse_db_entry()
            self._parse_server()

    def _parse_server(self):
        self._form["Computer name"] = self._job.server._host
        self._form["Cores"] = str(self._job.server.cores)

    def _parse_db_entry(self):
        db = self._job.database_entry
        if bool(db):
            self._form["Status"] = db.status
            self._form["Submission time"] = db.timestart
            self._form["Stop time"] = db.timestop
            self._form["Simulation type"] = db.hamilton
            self._form["Runtime [h]"] = str(db.totalcputime / 3600.0)
        else:
            self._form["Status"] = self._job["status"]

    @property
    def form(self):
        return self._form

    @classmethod
    def extract(cls, job, form):
        self = cls(job, form)
        return self.form


class CoscineResource(StorageInterface):
    def __init__(
        self,
        resource: Union[coscine.Resource, dict, "CoscineResource"],
        parent_path=None,
    ):
        """Giving access to a CoScInE Resource to receive or upload files

        To upload a file you need the file and a metadata form (a template is available using `metadata_template`,
        fill out the metadata and upload using the `upload_file` method
        ```
        coscine_res = CoscineResource(connection_dict)
        mdf = coscine_res.metadata_template
        mdf['ID'] = 'test'
        mdf['User'] = 'Niklas'
        mdf['Date'] = '11.04.2023'
        coscine_res.upload_file(file='some.txt', metadata=mdf)
        ```

        A pyiron atomistic simulation job can be uploaded simply by
        ```
        coscine_res.upload_job(job_object)
        ```

        If you want to receive a file from the storage, simply index the coscine resource with the name
        ```
        file_object = coscine_res['file_name_on_coscine']
        ```
        The return object has a metadata and a data attribute holding the respective information.

        Args:
            resource (coscine.Resource/dict): Either directly provide a coscine.Resource or a dictionary with
                token, project_id, and resource_id to directly connect to the respective resource.
        """
        super().__init__()
        if isinstance(resource, coscine.Resource):
            self._resource = resource
        elif isinstance(resource, dict):
            token = resource.pop("token", None)
            client, _ = CoscineConnect.get_client_and_object(token)
            pr = _list_filter(
                client.projects(toplevel=False), id=resource["project_id"]
            )[0]
            self._resource = _list_filter(pr.resources(), id=resource["resource_id"])[0]
        elif isinstance(resource, self.__class__):
            self._resource = resource._resource
            self._path = resource._path
            return
        else:
            raise TypeError(f"Unknown resource type {type(resource)}!")

        self._path = self._construct_path(parent_path)

    def _construct_path(self, parent_path):
        if parent_path is None:
            return (
                self._resource.project.display_name + "/" + self._resource.display_name
            )
        elif isinstance(parent_path, str) and parent_path.endswith(
            self._resource.project.display_name
        ):
            return parent_path + "/" + self._resource.display_name
        elif (
            isinstance(parent_path, list)
            and len(parent_path) > 0
            and parent_path[-1] == self._resource.project.display_name
        ):
            return "/".join(parent_path + [self._resource.display_name])
        else:
            state.logger.warn(
                "Provided parent path does not contain the parent. Fallback to Project_name/Resource_name"
            )
            return (
                self._resource.project.display_name + "/" + self._resource.display_name
            )

    @property
    def connection_info(self):
        return {
            "project_id": self._resource.project.id,
            "resource_id": self._resource.id,
        }

    def __repr__(self):
        """
        Human readable string representation of the project object

        Returns:
            str: string representation
        """
        return str(self.list_all())

    def _list_nodes(self):
        return [obj.name for obj in self._resource.objects()]

    def _list_groups(self):
        # Right now a coscine resource is flat.
        return []

    def is_dir(self, item):
        return False

    def is_file(self, item):
        if item in self.list_nodes():
            return True
        else:
            return False

    @property
    def resource(self) -> coscine.Resource:
        return self._resource

    @property
    def resource_type(self):
        return self.resource.profile

    def remove_file(self, item):
        file_obj = self._get_one_file_obj(
            item,
            error_msg=f"Multiple matches found for item = {item}, aborting remove!",
        )
        if file_obj is not None:
            file_obj.delete()

    def upload_file(
        self, file, metadata: coscine.resource.MetadataForm = None, filename=None
    ):
        _meta_data = self.validate_metadata(metadata)
        filename = filename or os.path.basename(file)
        self._resource.upload(filename, file, _meta_data)

    def upload_job(
        self,
        job: pyiron_base.GenericJob,
        form=None,
        update_form=False,
        dois=None,
        upload_working_dir=False,
    ):
        """Upload a pyiron job to this CoScInE resource

        Args:
            job: job object from pyiron
            form: optional metadata form, required if the metadata mapping between job and resource type is unknown.
            update_form(bool): If true update given form, else use as is.
            dois(str): Optional DOI of papers from this data, overwrites doi in form!
            upload_working_dir(bool): It true, compress and upload working directory.
        """
        job_file = job.project_hdf5.file_name

        if form is None:
            form = Job2CoscineMetadataConverter.extract(job, self.metadata_template)
        elif update_form:
            form = Job2CoscineMetadataConverter.extract(job, form)

        if dois is not None:
            form["DOIs"] = dois

        self.upload_file(job_file, form)
        if upload_working_dir:
            if not job.is_compressed():
                job.compress()
            self.upload_file(
                os.path.join(job.working_directory, job.job_name + ".tar.bz2"), form
            )

    def load_job(self, project, name: str, download_working_dir=False):
        """Load job using the provided project instance.

        Careful, this job is not saved, and if it is, it would not be in the database!

        Args:
            project: a pyiron Project instance
            name: the name of the job to load
        """

        if name.endswith(".h5"):
            job_name = name[:-3]
        elif name.endswith(".tar.bz2"):
            job_name = name[:-8]
        else:
            job_name = name

        hdf_name = job_name + ".h5"
        tar_name = job_name + ".tar.bz2"
        if hdf_name not in self.list_nodes():
            raise ValueError(f"No such job {job_name} found.")

        file = self[hdf_name]
        file_data = file.data
        if isinstance(file._data, str):
            path = "/" + file_data.keys()[0]
            job = ProjectHDFio(project, file._data, path).to_object()
            if download_working_dir and tar_name in self.list_nodes():
                job_file = file._data
                wd = os.path.dirname(file._data)
                tar_file = self[tar_name]
                tar_file.download(wd)
                os.rename(
                    os.path.join(wd, tar_name),
                    os.path.join(wd, job_file.replace(".h5", ".tar.bz2")),
                )
            return job
        else:
            raise ValueError(
                f"{name} does not seem to be a pyiron job or job related file!"
            )

    def _get_one_file_obj(self, item, error_msg) -> Union[coscine.FileObject, None]:
        if item not in self.list_nodes():
            return None

        objects = _list_filter(self._resource.objects(), name=item)
        if len(objects) == 1:
            return objects[0]
        elif len(objects) > 1:
            raise ValueError(error_msg)
        else:
            raise RuntimeError(
                f"This should be impossible. Please contact the developers."
            )

    def __getitem__(self, item):
        if item in self._list_nodes():
            return CoscineFileData(
                self._get_one_file_obj(
                    item,
                    f"Multiple matches for item = {item}, use `get_items(name={item})` to receive all matching files."
                    f" See get_items docstring for more filtering options.",
                )
            )
        raise ValueError(f"Unknown item {item}")

    def get_items(self, name=None, **kwargs) -> List[CoscineFileData]:
        """Use the coscine filtering options to receive all matching files.

        Args:
            name(str/None): The name of the file as provided by `list_nodes`. None is default.
            **kwargs: Additional keyword arguments as known to coscine. The most important of these is
                'Name' for which this method has the 'name' alias. Any file _not_ matching the value for the
                keyword will be excluded.
                In coscine-0.5.1 the following list of kwargs is available:
                ['Name', 'Path', 'Size', 'Kind', 'Modified', 'Created', 'Provider', 'IsFolder', 'IsFile', 'Action']
        Raises:
            ValueError: if 'name' and 'Name' are both given and do not match.
        """
        _name = kwargs.pop("Name", None)
        if name != _name:
            raise ValueError(
                f"name='{name}' and Name='{_name}' both provided and not matching."
            )

        kwargs["Name"] = name or _name
        return [
            CoscineFileData(obj)
            for obj in _list_filter(self._resource.objects(), **kwargs)
        ]

    @property
    def metadata_template(self) -> CoscineMetadata:
        return CoscineMetadata(self._resource.metadata_form())

    @property
    def requires_metadata(self):
        return True

    def _get_form_from_dict(self, metadata_dict):
        form = self.metadata_template
        form.clear()

        # Try to update form by the dictionary:
        try:
            form.fill(metadata_dict)
        except KeyError as e:
            store_err = KeyError(f"Meta data template has no key '{e.args[0]}'")
        else:
            return form, None

        # Assume we already got a dictionary from form.generate()
        # A form has the method parse(data) which is intended to be used on a coscine.Object.metadata which is the same
        # as {'file_handle_with_file_name': from.generate()} of the form used to upload; thus:
        form.parse({"dummy_file_name": metadata_dict})
        return form, store_err

    def parse_metadata(self, metadata):
        generated_metadata = self.validate_metadata(metadata, raise_error=False)
        if generated_metadata is None:
            raise ValueError("metadata is not valid")
        form, _ = self._get_form_from_dict(generated_metadata)
        return {key: val.raw() for key, val in form.items()}

    def validate_metadata(self, metadata, raise_error=True):
        encountered_error = None
        if metadata is None and raise_error:
            raise ValueError(
                "Coscine resources require meta data to upload a file! Use metadata_template to "
                "get the correct meta data form and provide a completed form here."
            )
        elif metadata is None:
            return None
        elif isinstance(metadata, coscine.resource.MetadataForm):
            if not isinstance(metadata, CoscineMetadata):
                meta = CoscineMetadata(metadata)
            else:
                meta = metadata
            metadata_form = self.metadata_template
            metadata_form.clear()
            metadata_form.fill(meta.to_dict())
        else:
            metadata_form, encountered_error = self._get_form_from_dict(metadata)

        try:
            result_meta_data = metadata_form.generate()
        except ValueError as e:
            if raise_error and encountered_error is not None:
                raise ValueError(
                    f"The provided meta data is not valid; might be related to previous error "
                    f"{encountered_error}"
                ) from e
            elif raise_error:
                raise e
            else:
                return None
        else:
            return result_meta_data


class CoscineConnect:
    def __init__(
        self,
        token: Union[
            coscine.Project,
            coscine.FileObject,
            coscine.Resource,
            coscine.Client,
            str,
            None,
        ],
    ):
        """
        project(coscine.project/coscine.client/str/None):
        """
        self._object = None
        if token is None:
            try:
                token = state.settings.credentials["COSCINE"]["TOKEN"]
            except (KeyError, AttributeError):
                token = getpass(prompt="Coscine token: ")
            self._client = self._connect_client(token)
        if isinstance(token, str) and os.path.isfile(token):
            with open(token) as f:
                token = f.read()
            self._client = self._connect_client(token)
        elif isinstance(token, str):
            self._client = self._connect_client(token)
        elif hasattr(token, "client"):
            self._object = token
            self._client = token.client
        else:
            self._client = token

        maintenance = self._client.get_maintenance()
        if maintenance["displayName"] is not None:
            state.logger.warn(f"Coscine is a maintenance mode: {maintenance}")

    @staticmethod
    def _connect_client(token: str):
        client = coscine.Client(token)
        try:
            client.projects()
        except (PermissionError, RuntimeError, ConnectionError) as e:
            raise ValueError("Error connecting to CoScInE with provided token.") from e
        return client

    @classmethod
    def get_client_and_object(cls, token):
        self = cls(token)
        return self._client, self._object


class CoscineProject(HasGroups):
    def __init__(
        self,
        project: Union[coscine.Project, coscine.Client, str, None] = None,
        parent_path=None,
    ):
        """
        project(coscine.project/coscine.client/str/None):
        parent_path
        """
        if parent_path is None:
            parent_path = []
        elif isinstance(parent_path, str):
            parent_path = [parent_path]
        self._path = None
        self._client, self._project = CoscineConnect.get_client_and_object(project)
        if self._project is not None:
            self._path = parent_path + [project.display_name]

    @property
    def path(self):
        """A convenience path representation (human readable)"""
        if self._path is None:
            return ""
        return "/".join(self._path)

    @property
    def read_only(self):
        return self._client.settings.read_only

    @read_only.setter
    def read_only(self, val: bool):
        self._client.settings.read_only = val

    @property
    def verbose(self):
        return self._client.settings.verbose

    def __repr__(self):
        """
        Human readable string representation of the project object

        Returns:
            str: string representation
        """
        return str(self.list_all())

    @verbose.setter
    def verbose(self, val):
        self._client.settings.verbose = val

    def _list_groups(self):
        if self._project is None:
            return [pr.name for pr in self._client.projects()]
        else:
            return [pr.name for pr in self._project.subprojects()]

    def _list_nodes(self):
        if self._project is None:
            return []
        else:
            return [res.name for res in self._project.resources()]

    def __getitem__(self, key):
        if key in self.list_nodes():  # This implies project is not None
            return CoscineResource(self._project.resource(key), self._path)
        return self.get_group(key)

    def get_node(self, key):
        if key in self.list_nodes():
            return CoscineResource(self._project.resource(key), self._path)
        else:
            return KeyError(key)

    def get_group(self, key):
        if key in self.list_groups() and self._project is not None:
            try:
                return self.__class__(
                    self._project.subproject(display_name=key), parent_path=self._path
                )
            except IndexError:
                warnings.warn("More than one project matches - returning first match!")
                for _pr in self._project.subprojects():
                    if _pr.display_name == key:
                        return self.__class__(_pr)
        elif key in self.list_groups():
            return self.__class__(self._client.project(key))
        else:
            raise KeyError(key)

    def create_node(
        self,
        name_or_form,
        size=None,
        display_name=None,
        description=None,
        disciplines=None,
        keywords=None,
        metadata_visibility=None,
        licence=None,
        internal_rules_for_reuse=None,
        application_profile="sfb1394/AtomisticSimulation",
        resource_type="rdsrwth",
    ):
        """Creates a new coscine resource

        Args:
            name_or_form(str/form): Name of the resource or coscine resource form to construct the resource
                                    The form has to be generated by the coscine client.
            size(int): size of the resource in GB
            display_name(str): Name of the resource displayed on the web, optional: taking name if not provided
            description(str): Description of the resource, optional: taking name if not provided
            disciplines([str]): List of disciplines as known to coscine, optional: taking form project if not provided
            keywords(str): Keywords describing the resource
            metadata_visibility(str): For whom the metadata is visible,
                                      optional: taking from the project if not provided
                                      choices: "Project Members", "Public"
            licence(str): Licence for the resource as known to coscine
            internal_rules_for_reuse(str):
            application_profile(str): What metadata scheme to use, default: "sfb1394/AtomisticSimulation"
            resource_type(str): What type of resource is created, default: "rdsrwth"
        """
        if self._project is None:
            raise RuntimeError(
                "At the top level, new resources cannot be created. Switch to a project"
            )
        if not isinstance(name_or_form, str):
            form = name_or_form
        else:
            form = self._client.resource_form()
            form["Resource Type"] = resource_type
            form["Resource Size"] = size or 1
            form["Resource Name"] = name_or_form
            form["Display Name"] = display_name or name_or_form
            form["Resource Description"] = description or name_or_form
            form["Discipline"] = disciplines or self._project.disciplines
            if keywords:
                form["Resource Keywords"] = keywords
            form["Metadata Visibility"] = (
                metadata_visibility or self._project.visibility
            )
            if licence:
                form["License"] = licence
            if internal_rules_for_reuse:
                form["Internal Rules for Reuse"] = internal_rules_for_reuse
            form["Application Profile"] = application_profile
        if form["Resource Name"] in self.list_groups() + self.list_nodes():
            raise ValueError("This name is already in this project!")
        return CoscineResource(
            self._project.create_resource(form), parent_path=self.path
        )

    def create_group(
        self,
        project_name,
        display_name=None,
        project_description=None,
        principal_investigators=None,
        project_start=None,
        project_end=None,
        discipline=None,
        participating_organizations=None,
        project_keywords=None,
        metadata_visibility=None,
        grant_id=None,
        copy_members=True,
    ):
        if project_name in self.list_nodes() + self.list_groups():
            raise ValueError("The name is already in this project!")
        update_dict = {
            "Project Name": project_name,
            "Display Name": display_name,
            "Project Description": project_description,
            "Principal Investigators": principal_investigators,
            "Project Start": project_start,
            "Project End": project_end,
            "Discipline": discipline,
            "Participating Organizations": participating_organizations,
            "Project Keywords": project_keywords,
            "Metadata Visibility": metadata_visibility,
            "Grant ID": grant_id,
        }
        for key, val in update_dict.copy().items():
            if val is None:
                del update_dict[key]

        if "Display Name" not in update_dict:
            update_dict["Display Name"] = project_name

        for key in ["Project Start", "Project End"]:
            if key in update_dict and isinstance(update_dict[key], str):
                update_dict[key] = parser.parse(update_dict[key])

        form = (
            self._client.project_form()
            if self._project is None
            else self._project.form()
        )
        form.fill(update_dict)
        form.generate()
        if self._project is None:
            new = self._client.create_project(form)
        else:
            new = self._project.create_subproject(form)
            if copy_members:
                with coscine.concurrent():
                    for member in self._project.members():
                        new.add_member(member, role=member.role)

        return new



import xml.etree.ElementTree as ET
import copy

import numpy as np
import matplotlib.pyplot as plt

from pyiron_base import PyironFactory, DataContainer


class FunctionFactory(PyironFactory):
    """
    Class to conveniently create different function objects.
    for detailed information about the function visit the
    atomicrex documentation.
    """

    @staticmethod
    def user_function(
        identifier,
        input_variable="r",
        species=["*", "*"],
        is_screening_function=False,
        cutoff=None,
    ):
        return UserFunction(
            identifier,
            input_variable=input_variable,
            species=species,
            is_screening_function=is_screening_function,
            cutoff=cutoff,
        )

    @staticmethod
    def poly(identifier, cutoff, species=["*", "*"]):
        """
        TAKE CARE !!!
        The polynomial function implemented in atomicrex does not handle derivatives at the cutoff right now,
        i.e. when using this as a pair function or similar there will be massive jumps in forces.
        """
        return Poly(identifier, cutoff=cutoff, species=species)

    @staticmethod
    def spline(
        identifier, cutoff, derivative_left=0, derivative_right=0, species=["*", "*"]
    ):
        return Spline(identifier, cutoff, derivative_left, derivative_right, species)

    @staticmethod
    def equidistant_spline(
        identifier,
        n_nodes,
        cutoff,
        initial_value_func,
        min_distance=0.0,
        derivative_left=0.0,
        d_left_enabled=True,
        derivative_right=0.0,
        d_right_enabled=False,
        endpoint_val=0.0,
        species=["*", "*"],
    ):
        """
        Convenience function to create a spline function with equidistant node points.

        Args:
            identifier (str): function identifier. Should be unique within the job
            n_nodes (int): number of node points
            cutoff (float): values after are 0
            initial_value_func (function(x)): function to calculate start values for nodes.
            min_distance (float, optional): x coordinate of first node point. Defaults to 0.
            derivative_left (int, optional): . Defaults to 0.
            d_left_enabled (bool, optional): Whether to fit. Defaults to True.
            derivative_right (int, optional): [description]. Defaults to 0.
            d_right_enabled (bool, optional): Whether to fit. Should be False for most functions, beside Embedding terms. Defaults to False.
            endpoint_val (float, None, bool, optional): Start val for endpoint, enabled=False if float, if False endpoint is not included. If None endpoint is included, enabled and start_val is calculated like other points. Defaults to 0.0, which should be used in most cases.
            species (list of str, optional): Only needs to be changed for multi element fits. Defaults to ["*", "*"].

        Returns:
            [type]: [description]
        """
        s = Spline(identifier, cutoff, derivative_left, derivative_right, species)
        if endpoint_val is False:
            x = np.linspace(
                start=min_distance, stop=cutoff, num=n_nodes, endpoint=False
            )
        else:
            x = np.linspace(start=min_distance, stop=cutoff, num=n_nodes, endpoint=True)
        y = initial_value_func(x)
        s.parameters.create_from_arrays(x, y)
        if endpoint_val is not False and endpoint_val is not None:
            s.parameters[f"node_{cutoff}"].enabled = False
            s.parameters[f"node_{cutoff}"].start_val = endpoint_val

        s.derivative_left.enabled = d_left_enabled
        s.derivative_right.enabled = d_right_enabled
        s.derivative_left.start_val = derivative_left
        s.derivative_right.start_val = derivative_right
        return s

    @staticmethod
    def exp_A_screening(
        identifier, cutoff, species=["*", "*"], is_screening_function=True
    ):
        return ExpA(
            identifier,
            cutoff,
            species=species,
            is_screening_function=is_screening_function,
        )

    @staticmethod
    def exp_B_screening(
        identifier,
        cutoff,
        rc,
        alpha,
        exponent,
        species=["*", "*"],
        is_screening_function=True,
    ):
        return ExpB(
            identifier,
            cutoff,
            rc,
            alpha,
            exponent,
            species=species,
            is_screening_function=is_screening_function,
        )

    @staticmethod
    def exp_gaussian_screening(
        identifier,
        cutoff,
        stddev,
        alpha,
        exponent,
        species=["*", "*"],
        is_screening_function=True,
    ):
        return ExpGaussian(
            identifier,
            cutoff,
            stddev,
            alpha,
            exponent,
            species=species,
            is_screening_function=is_screening_function,
        )

    @staticmethod
    def morse_A(identifier, D0, r0, alpha, species=["*", "*"]):
        return MorseA(identifier, D0, r0, alpha, species=species)

    @staticmethod
    def morse_B(identifier, D0, r0, beta, S, delta, species=["*", "*"]):
        return MorseB(identifier, D0, r0, beta, S, delta, species=species)

    @staticmethod
    def morse_C(identifier, A, B, mu, lambda_val, delta, species=["*", "*"]):
        return MorseC(identifier, A, B, mu, lambda_val, delta, species=species)

    @staticmethod
    def gaussian(identifier, prefactor, eta, mu, species=["*", "*"], cutoff=None):
        return GaussianFunc(identifier, prefactor, eta, mu, species, cutoff)

    @staticmethod
    def x_pow_n_cutoff(
        identifier, cutoff, h=1, N=4, species=["*", "*"], is_screening_function=True
    ):
        return XpowNCutoff(
            identifier=identifier,
            cutoff=cutoff,
            h=h,
            N=N,
            species=species,
            is_screening_function=is_screening_function,
        )

    @staticmethod
    def constant(identifier, constant, species=["*", "*"]):
        return Constant(constant=constant, identifier=identifier, species=species)

    @staticmethod
    def MishinCuV(
        identifier,
        E1,
        E2,
        alpha1,
        alpha2,
        r01,
        r02,
        delta,
        cutoff,
        h,
        S1,
        rs1,
        S2,
        rs2,
        S3,
        rs3,
        species=["*", "*"],
    ):
        product_func = FunctionFactory.product(identifier, species)
        sum_func = FunctionFactory.sum(identifier="MorseSum", species=species)
        morse1 = FunctionFactory.morse_A(
            identifier="Morse1", D0=E1, r0=r01, alpha=alpha1, species=species
        )
        morse2 = FunctionFactory.morse_A(
            identifier="Morse2", D0=E2, r0=r02, alpha=alpha2, species=species
        )
        c = FunctionFactory.constant(
            identifier="delta", constant=delta, species=species
        )
        rep1 = FunctionFactory.RsMinusRPowN(
            identifier="rep1", S=S1, rs=rs1, N=4, species=species
        )
        rep2 = FunctionFactory.RsMinusRPowN(
            identifier="rep2", S=S2, rs=rs2, N=4, species=species
        )
        rep3 = FunctionFactory.RsMinusRPowN(
            identifier="rep3", S=S3, rs=rs3, N=4, species=species
        )
        rep1.parameters.N.enabled = False
        rep2.parameters.N.enabled = False
        rep3.parameters.N.enabled = False
        sum_func.functions[morse1.identifier] = morse1
        sum_func.functions[morse2.identifier] = morse2
        sum_func.functions[c.identifier] = c
        sum_func.functions[rep1.identifier] = rep1
        sum_func.functions[rep2.identifier] = rep2
        sum_func.functions[rep3.identifier] = rep3
        screening = FunctionFactory.x_pow_n_cutoff(
            identifier="screening", cutoff=cutoff, h=h, N=4, species=species
        )
        screening.is_screening_function = False
        screening.screening = None
        product_func.functions[sum_func.identifier] = sum_func
        product_func.functions[screening.identifier] = screening
        return product_func

    @staticmethod
    def MishinCuRho(identifier, a, r1, r2, beta1, beta2, species=["*", "*"]):
        return MishinCuRho(identifier, a, r1, r2, beta1, beta2, species)

    @staticmethod
    def MishinCuF(identifier, F0, F2, q1, q2, q3, q4, Q1, Q2, species=["*"]):
        return MishinCuF(identifier, F0, F2, q1, q2, q3, q4, Q1, Q2, species)

    @staticmethod
    def extendedMishinCuF(
        identifier, F0, F2, f3, f4, f5, f6, a3, a4, a5, a6, d3, d4, d5, species=["*"]
    ):
        return ExtendedMishinCuF(
            identifier=identifier,
            F0=F0,
            F2=F2,
            f3=f3,
            f4=f4,
            f5=f5,
            f6=f6,
            a3=a3,
            a4=a4,
            a5=a5,
            a6=a6,
            d3=d3,
            d4=d4,
            d5=d5,
            species=species,
        )

    @staticmethod
    def RsMinusRPowN(identifier, S, rs, N, species=["*", "*"], cutoff=None):
        return RsMinusRPowN(identifier, S, rs, N, species, cutoff=cutoff)

    @staticmethod
    def sum(identifier, species=["*", "*"]):
        return Sum(identifier=identifier, species=species)

    @staticmethod
    def product(identifier, species=["*", "*"]):
        return Product(identifier=identifier, species=species)

    @staticmethod
    def gaussians_sum(
        n_gaussians,
        eta,
        identifier,
        node_points=None,
        cutoff=None,
        initial_prefactors=None,
        min_prefactors=None,
        max_prefactors=None,
        species=["*", "*"],
    ):
        sum_func = FunctionFactory.sum(identifier=identifier, species=species)
        if node_points is None:
            if cutoff is None:
                raise ValueError(
                    "Specify node points or a cutoff to set them automatically"
                )
            else:
                node_points = np.linspace(0, cutoff, n_gaussians, endpoint=False)
        else:
            if len(node_points) != n_gaussians:
                raise ValueError("Number of node points has to match n_gaussians")

        if initial_prefactors is None:
            initial_prefactors = np.ones(n_gaussians)
        if min_prefactors is not None:
            if len(min_prefactors) != n_gaussians:
                raise ValueError("min_prefactors must have length num_gaussians")
        if max_prefactors is not None:
            if len(max_prefactors) != n_gaussians:
                raise ValueError("max_prefactors must have length num_gaussians")

        for i in range(n_gaussians):
            gauss = FunctionFactory.gaussian(
                identifier=f"gauss_{i}",
                prefactor=initial_prefactors[i],
                eta=eta,
                mu=node_points[i],
                species=species,
                cutoff=cutoff,
            )
            gauss.parameters.mu.enabled = False
            gauss.parameters.eta.enabled = False
            if min_prefactors is not None:
                gauss.parameters.prefactor.min_val = min_prefactors[i]
            if max_prefactors is not None:
                gauss.parameters.prefactor.max_val = max_prefactors[i]

            sum_func.functions[gauss.identifier] = gauss
        return sum_func


class BaseFunctionMixin:
    # Mixin class to implement functionality common in all types of functions
    # Be careful with Spline class because it has params, but also derivatives, requiring some special additions to implementations
    def copy_final_to_initial_params(self, filter_func=None):
        for param in self.parameters.values():
            param.copy_final_to_start_value(filter_func=filter_func)

    def lock_parameters(self, filter_func=None):
        for param in self.parameters.values():
            param.lock(filter_func=filter_func)

    def set_max_values(self, constant=None, factor=None, filter_func=None):
        """
        Convenience function so set max values for all parameters at once.
        Can either use a constant value or a factor. If both are given factor is used.

        Args:
            constant ([type], optional): param.max_val = constant. Defaults to None.
            factor ([type], optional): param.max_val = abs(start_val)*factor. Defaults to None.
            filter_func ([type], optional): Optional function to filter params. Should take param as argument and return True or False. Defaults to None.

        Raises:
            ValueError: Raises when constant and factor are None.
        """
        if constant is None and factor is None:
            raise ValueError("constant or factor must be set")

        for param in self.parameters.values():
            param.set_max_val(
                constant=constant,
                factor=factor,
                filter_func=filter_func,
            )

    def set_min_values(self, constant=None, factor=None, filter_func=None):
        """
        Convenience function so set min values for all parameters at once.
        Can either use a constant value or a factor. If both are given factor is used.

        Args:
            constant ([type], optional): param.min_val = constant. Defaults to None.
            factor ([type], optional): param.min_val = -abs(start_val)*factor. Defaults to None.
            filter_func ([type], optional): Optional function to filter params. Should take param as argument and return True or False. Defaults to None.

        Raises:
            ValueError: Raises when constant and factor are None.
        """
        if constant is None and factor is None:
            raise ValueError("constant or factor must be set")

        for param in self.parameters.values():
            param.set_min_val(
                constant=constant,
                factor=factor,
                filter_func=filter_func,
            )

    def count_parameters(self, enabled_only=True):
        parameters = 0
        if enabled_only:
            for param in self.parameters.values():
                if param.enabled:
                    parameters += 1
        else:
            for param in self.parameters.values():
                parameters += 1
        return parameters


class MetaFunctionMixin:
    def copy_final_to_initial_params(self, filter_func=None):
        for f in self.functions.values():
            f.copy_final_to_initial_params(filter_func=filter_func)

    def lock_parameters(self, filter_func=None):
        for f in self.functions.values():
            f.lock_parameters(filter_func=filter_func)

    def set_max_values(self, constant=None, factor=None, filter_func=None):
        """
        Convenience function so set max values for all parameters at once.
        Can either use a constant value or a factor. If both are given factor is used.

        Args:
            constant ([type], optional): param.max_val = constant. Defaults to None.
            factor ([type], optional): param.max_val = abs(start_val)*factor. Defaults to None.
            filter_func ([type], optional): Optional function to filter params. Should take param as argument and return True or False. Defaults to None.

        Raises:
            ValueError: Raises when constant and factor are None.
        """
        for f in self.functions.values():
            f.set_max_values(constant=constant, factor=factor, filter_func=filter_func)

    def set_min_values(self, constant=None, factor=None, filter_func=None):
        """
        Convenience function so set min values for all parameters at once.
        Can either use a constant value or a factor. If both are given factor is used.

        Args:
            constant ([type], optional): param.min_val = constant. Defaults to None.
            factor ([type], optional): param.min_val = -abs(start_val)*factor. Defaults to None.
            filter_func ([type], optional): Optional function to filter params. Should take param as argument and return True or False. Defaults to None.

        Raises:
            ValueError: Raises when constant and factor are None.
        """
        for f in self.functions.values():
            f.set_min_values(constant=constant, factor=factor, filter_func=filter_func)

    def count_parameters(self, enabled_only=True):
        parameters = 0
        for f in self.functions.values():
            parameters += f.count_parameters(enabled_only=enabled_only)
        return parameters


class AbstractMetaFunction(DataContainer, MetaFunctionMixin):
    def __init__(self, identifier=None, species=None, table_name=None):
        super().__init__()
        self.identifier = identifier
        self.functions = DataContainer(table_name=table_name)
        self.species = species

    def _to_xml_element(self, func_name):
        root = ET.Element(func_name)
        root.set("id", self.identifier)
        for k, v in self.functions.items():
            root.append(v._to_xml_element())
        return root

    def _parse_final_parameter(self, leftover, value):
        identifier = leftover[0].split("[")[0]
        leftover = leftover[1:]
        try:
            self.functions[identifier]._parse_final_parameter(leftover, value)
        except KeyError:
            raise KeyError(f"Function {identifier} not found in {self.identifier}")


class Sum(AbstractMetaFunction, MetaFunctionMixin):
    def __init__(self, identifier=None, species=None):
        super().__init__(
            identifier=identifier, species=species, table_name="sum_functions"
        )

    def _to_xml_element(self):
        return super()._to_xml_element(func_name="sum")


class Product(AbstractMetaFunction, MetaFunctionMixin):
    def __init__(self, identifier=None, species=None):
        super().__init__(
            identifier=identifier, species=species, table_name="product_functions"
        )

    def _to_xml_element(self):
        return super()._to_xml_element(func_name="product")


class SpecialFunction(DataContainer, BaseFunctionMixin):
    """
    Analytic functions defined within atomicrex should inherit from this class
    https://atomicrex.org/potentials/functions.html#index-1
    https://atomicrex.org/potentials/functions.html#specialized-functions
    """

    def __init__(
        self, identifier=None, species=["*", "*"], is_screening_function=False
    ):
        super().__init__(table_name=f"special_function_{identifier}")
        self.species = species
        self.parameters = FunctionParameterList()
        self.is_screening_function = is_screening_function
        self.identifier = identifier
        if not is_screening_function:
            self.screening = None

    def _to_xml_element(self, name):
        if self.is_screening_function:
            screening = ET.Element("screening")
            root = ET.SubElement(screening, f"{name}")
        else:
            root = ET.Element(f"{name}")

        root.set("id", f"{self.identifier}")
        for param in self.parameters.values():
            p = ET.SubElement(root, f"{param.param}")
            p.text = f"{param.start_val}"

        # This if condition is to prevent an error with the expA screening function
        if name != "exp-A":
            root.append(self.parameters.fit_dofs_to_xml_element())

        if not self.is_screening_function:
            if self.screening is not None:
                root.append(self.screening._to_xml_element())
            return root
        else:
            return screening

    @property
    def func(self):
        return None

    def plot(self):
        if self.func is None:
            raise NotImplementedError(
                "A func property needs to be defined in the subclass"
            )
        else:
            return plot(self.func)

    def _parse_final_parameter(self, leftover, value):
        param = leftover[0].rstrip(":")
        self.parameters[param].final_value = value


class Poly(DataContainer, BaseFunctionMixin):
    """
    Polynomial interpolation function.
    """

    def __init__(self, identifier=None, cutoff=None, species=["*", "*"]):
        super().__init__(table_name=f"Poly_{identifier}")
        self.identifier = identifier
        self.cutoff = cutoff
        self.species = species
        self.parameters = PolyCoeffList()
        # preparation if poly gets screening function ability
        # self.screening = None

    def _to_xml_element(self):
        poly = ET.Element("poly")
        poly.set("id", self.identifier)
        cutoff = ET.SubElement(poly, "cutoff")
        cutoff.text = f"{self.cutoff}"
        poly.append(self.parameters._to_xml_element())
        # if self.screening is not None:
        #        poly.append(self.screening._to_xml_element())
        return poly


class Spline(DataContainer, BaseFunctionMixin):
    """
    Spline interpolation function
    """

    def __init__(
        self,
        identifier=None,
        cutoff=None,
        derivative_left=0,
        derivative_right=0,
        species=["*", "*"],
    ):
        super().__init__(table_name=f"Spline_{identifier}")
        self.identifier = identifier
        self.cutoff = cutoff
        self.derivative_left = FunctionParameter(
            param="derivative-left", start_val=derivative_left
        )
        self.derivative_right = FunctionParameter(
            param="derivative-right", start_val=derivative_right, enabled=False
        )
        self.species = species
        self.parameters = NodeList()

    def _to_xml_element(self):
        spline = ET.Element("spline")
        spline.set("id", self.identifier)
        if self.cutoff is not None:
            cutoff = ET.SubElement(spline, "cutoff")
            cutoff.text = f"{self.cutoff}"
        der_l = ET.SubElement(spline, "derivative-left")
        der_l.text = f"{self.derivative_left.start_val}"
        der_r = ET.SubElement(spline, "derivative-right")
        der_r.text = f"{self.derivative_right.start_val}"

        fit_dof = ET.SubElement(spline, "fit-dof")
        fit_dof.append(self.derivative_left._to_xml_element())
        fit_dof.append(self.derivative_right._to_xml_element())

        spline.append(self.parameters._to_xml_element())
        return spline

    def _parse_final_parameter(self, leftover, value):
        if "derivative-right" in leftover[-1]:
            self.derivative_right.final_value = value
        elif "derivative-left" in leftover[-1]:
            self.derivative_left.final_value = value
        else:
            param = float(leftover[0].split("[")[1])
            param = f"node_{param:.6g}"
            self.parameters[param].final_value = value

    def copy_final_to_initial_params(self, filter_func=None):
        super().copy_final_to_initial_params(filter_func=filter_func)
        self.derivative_left.copy_final_to_start_value(filter_func=filter_func)
        self.derivative_right.copy_final_to_start_value(filter_func=filter_func)

    def lock_parameters(self, filter_func=None):
        super().lock_parameters(filter_func=filter_func)
        self.derivative_left.lock(filter_func=filter_func)
        self.derivative_right.lock(filter_func=filter_func)

    def set_max_values(self, constant=None, factor=None, filter_func=None):
        super().set_max_values(constant, factor, filter_func)
        self.derivative_left.set_max_val(
            constant=constant, factor=factor, filter_func=filter_func
        )
        self.derivative_right.set_max_val(
            constant=constant, factor=factor, filter_func=filter_func
        )

    def set_min_values(self, constant=None, factor=None, filter_func=None):
        super().set_min_values(constant, factor, filter_func)
        self.derivative_left.set_min_val(
            constant=constant, factor=factor, filter_func=filter_func
        )
        self.derivative_right.set_min_val(
            constant=constant, factor=factor, filter_func=filter_func
        )

    def count_parameters(self, enabled_only=True):
        parameters = super().count_parameters(enabled_only=enabled_only)
        if enabled_only:
            if self.derivative_left.enabled:
                parameters += 1
            if self.derivative_right.enabled:
                parameters += 1
        else:
            parameters += 2
        return parameters


class ExpA(SpecialFunction):
    def __init__(
        self,
        identifier=None,
        cutoff=None,
        species=["*", "*"],
        is_screening_function=True,
    ):
        super().__init__(
            identifier, species=species, is_screening_function=is_screening_function
        )
        self.parameters.add_parameter(
            "cutoff",
            start_val=cutoff,
            enabled=False,
            fitable=False,
        )

    @property
    def func(self):
        return lambda r: np.exp(1 / (r - self.parameters.cutoff.start_val))

    def _to_xml_element(self):
        return super()._to_xml_element(name="exp-A")


class ExpB(SpecialFunction):
    def __init__(
        self,
        identifier=None,
        cutoff=None,
        rc=None,
        alpha=None,
        exponent=None,
        species=None,
        is_screening_function=True,
    ):
        super().__init__(
            identifier, species=species, is_screening_function=is_screening_function
        )
        self.parameters.add_parameter(
            "cutoff",
            start_val=cutoff,
            enabled=False,
            fitable=False,
        )
        self.parameters.add_parameter(
            "rc",
            start_val=rc,
            enabled=False,
        )
        self.parameters.add_parameter(
            "alpha",
            start_val=alpha,
            enabled=False,
        )
        self.parameters.add_parameter(
            "exponent",
            start_val=exponent,
            enabled=False,
        )

    @property
    def func(self):
        return lambda r: np.exp(
            -np.sign(self.parameters.exponent.start_val)
            * self.parameters.alpha.start_val
            / (
                1
                - (
                    (r - self.parameters.rc.start_val)
                    / self.parameters.cutoff.start_val
                    - self.parameters.rc.start_val
                )
                ** self.parameters.exponent.start_val
            )
        )

    def _to_xml_element(self):
        return super()._to_xml_element(name="exp-B")


class ExpGaussian(SpecialFunction):
    def __init__(
        self,
        identifier=None,
        cutoff=None,
        stddev=None,
        alpha=None,
        exponent=None,
        species=["*", "*"],
        is_screening_function=True,
    ):
        super().__init__(
            identifier, species=species, is_screening_function=is_screening_function
        )
        self.parameters.add_parameter(
            "cutoff",
            start_val=cutoff,
            enabled=False,
            fitable=False,
        )
        self.parameters.add_parameter(
            "stddev",
            start_val=stddev,
            enabled=False,
        )
        self.parameters.add_parameter(
            "alpha",
            start_val=alpha,
            enabled=False,
        )
        self.parameters.add_parameter(
            "exponent",
            start_val=exponent,
            enabled=False,
        )

    @property
    def func(self):
        cutoff = self.parameters["cutoff"].start_val
        stddev = self.parameters["stddev"].start_val
        alpha = self.parameters["alpha"].start_val
        exponent = self.parameters["exponent"].start_val
        return (
            lambda r: np.exp(
                -np.sign(exponent) * alpha / (1 - (r / cutoff) ** exponent)
            )
            * np.exp(-(r**2) / (2 * stddev**2))
            / (stddev * np.sqrt(2 * np.pi))
        )

    def _to_xml_element(self):
        return super()._to_xml_element(name="exp-gaussian")


class XpowNCutoff(SpecialFunction):
    def __init__(
        self,
        identifier=None,
        cutoff=None,
        h=1,
        N=4,
        species=["*", "*"],
        is_screening_function=True,
    ):
        super().__init__(
            identifier, species=species, is_screening_function=is_screening_function
        )
        self.parameters.add_parameter(
            "cutoff",
            start_val=cutoff,
            enabled=False,
            fitable=False,
        )
        self.parameters.add_parameter(
            "h",
            start_val=h,
            enabled=False,
        )
        self.parameters.add_parameter(
            "N",
            start_val=N,
            enabled=False,
        )

    @property
    def func(self):
        rc = self.parameters.cutoff.start_val
        h = self.parameters.h.start_val
        N = self.parameters.N.start_val
        return lambda r: ((r - rc) / h) ** N / (1 + ((r - rc) / h) ** N)

    def _to_xml_element(self):
        return super()._to_xml_element(name="XpowN-cutoff")


class MorseA(SpecialFunction):
    def __init__(
        self, identifier=None, D0=None, r0=None, alpha=None, species=["*", "*"]
    ):
        super().__init__(identifier, species=species, is_screening_function=False)
        self.parameters.add_parameter(
            "D0",
            start_val=D0,
            enabled=True,
        )
        self.parameters.add_parameter(
            "r0",
            start_val=r0,
            enabled=True,
        )
        self.parameters.add_parameter(
            "alpha",
            start_val=alpha,
            enabled=True,
        )

    @property
    def func(self):
        return lambda r: self.parameters.D0.start_val * (
            np.exp(
                -2
                * self.parameters.alpha.start_val
                * (r - self.parameters.r0.start_val)
            )
            - 2
            * np.exp(
                -self.parameters.alpha.start_val * (r - self.parameters.r0.start_val)
            )
        )

    def _to_xml_element(self):
        return super()._to_xml_element(name="morse-A")


class MorseB(SpecialFunction):
    def __init__(
        self,
        identifier=None,
        D0=None,
        r0=None,
        beta=None,
        S=None,
        delta=None,
        species=["*", "*"],
    ):
        super().__init__(identifier, species=species, is_screening_function=False)
        self.parameters.add_parameter(
            "D0",
            start_val=D0,
            enabled=True,
        )
        self.parameters.add_parameter(
            "r0",
            start_val=r0,
            enabled=True,
        )
        self.parameters.add_parameter(
            "beta",
            start_val=beta,
            enabled=True,
        )
        self.parameters.add_parameter(
            "S",
            start_val=S,
            enabled=True,
        )
        self.parameters.add_parameter(
            "delta",
            start_val=delta,
            enabled=True,
        )

    @property
    def func(self):
        D0 = self.parameters.D0.start_val
        r0 = self.parameters.r0.start_val
        S = self.parameters.S.start_val
        beta = self.parameters.beta.start_val
        delta = self.parameters.delta.start_val
        return lambda r: (
            D0 / (S - 1) * np.exp(-beta * np.sqrt(2 * S) * (r - r0))
            - D0 * S / (S - 1) * np.exp(-beta * np.sqrt(2 / S) * (r - r0))
            + delta
        )

    def _to_xml_element(self):
        return super()._to_xml_element(name="morse-B")


class MorseC(SpecialFunction):
    def __init__(
        self,
        identifier=None,
        A=None,
        B=None,
        mu=None,
        lambda_val=None,
        delta=None,
        species=["*", "*"],
    ):
        super().__init__(identifier, species=species, is_screening_function=False)
        self.parameters.add_parameter(
            "A",
            start_val=A,
            enabled=True,
        )

        self.parameters.add_parameter(
            "B",
            start_val=B,
            enabled=True,
        )

        self.parameters.add_parameter(
            "mu",
            start_val=mu,
            enabled=True,
        )

        self.parameters.add_parameter(
            "lambda",
            start_val=lambda_val,
            enabled=True,
        )

        self.parameters.add_parameter(
            "delta",
            start_val=delta,
            enabled=True,
        )

    @property
    def func(self):
        A = self.parameters["A"].start_val
        B = self.parameters["B"].start_val
        mu = self.parameters["mu"].start_val
        param_lambda = self.parameters["lambda"].start_val
        delta = self.parameters["delta"].start_val
        return lambda r: A * np.exp(-param_lambda * r) - B * np.exp(-mu * r) + delta

    def _to_xml_element(self):
        return super()._to_xml_element(name="morse-C")


class RsMinusRPowN(SpecialFunction):
    def __init__(
        self,
        identifier=None,
        S=None,
        rs=None,
        N=None,
        species=None,
        is_screening_function=False,
        cutoff=None,
    ):
        super().__init__(
            identifier, species=species, is_screening_function=is_screening_function
        )
        self.cutoff = cutoff
        self.parameters.add_parameter(
            "S",
            start_val=S,
            enabled=True,
        )
        self.parameters.add_parameter(
            "rs",
            start_val=rs,
            enabled=True,
        )
        self.parameters.add_parameter(
            "N",
            start_val=N,
            enabled=False,
        )

    @property
    def func(self):
        def func(r):
            if r < self.parameters.rs.start_val:
                return (
                    self.parameters.S.start_val
                    * (self.parameters.rs.start_val - r) ** self.parameters.N.start_val
                )
            return 0

        return func

    def _to_xml_element(self):
        xml = super()._to_xml_element(name="RsMinusRPowN")
        if self.cutoff is not None:
            cutoff = ET.SubElement(xml, "cutoff")
            cutoff.text = f"{self.cutoff}"
        return xml


class Constant(SpecialFunction):
    def __init__(self, constant=None, identifier=None, species=["*", "*"]):
        super().__init__(identifier, species=species, is_screening_function=False)
        self.parameters.add_parameter(
            "const",
            start_val=constant,
            enabled=True,
        )

    def _to_xml_element(self):
        return super()._to_xml_element(name="constant")


## Renamed GaussianFunc to not mess with the gaussian code when loading from hdf5
class GaussianFunc(SpecialFunction):
    def __init__(
        self,
        identifier=None,
        prefactor=None,
        eta=None,
        mu=None,
        species=None,
        cutoff=None,
    ):
        super().__init__(identifier, species=species, is_screening_function=False)
        self.parameters.add_parameter(
            "prefactor",
            start_val=prefactor,
            enabled=True,
        )

        self.parameters.add_parameter(
            "eta",
            start_val=eta,
            enabled=True,
        )

        self.parameters.add_parameter(
            "mu",
            start_val=mu,
            enabled=True,
        )
        self.cutoff = cutoff

    @property
    def func(self):
        prefactor = self.parameters["prefactor"].start_val
        eta = self.parameters["eta"].start_val
        mu = self.parameters["mu"].start_val
        return lambda r: prefactor * np.exp(-eta * (r - mu) ** 2)

    def _to_xml_element(self):
        xml = super()._to_xml_element(name="gaussian")
        # Put this in SpecialFunction or AbstractFunction class when rewriting
        # f.e. using getattr()
        if self.cutoff is not None:
            cutoff = ET.SubElement(xml, "cutoff")
            cutoff.text = f"{self.cutoff}"
        return xml


class MishinCuRho(SpecialFunction):
    def __init__(
        self,
        identifier=None,
        a=None,
        r1=None,
        r2=None,
        beta1=None,
        beta2=None,
        species=["*", "*"],
    ):
        super().__init__(identifier, species=species, is_screening_function=False)
        self.parameters.add_parameter(
            "a",
            start_val=a,
            enabled=True,
        )
        self.parameters.add_parameter(
            "r1",
            start_val=r1,
            enabled=True,
        )
        self.parameters.add_parameter(
            "r2",
            start_val=r2,
            enabled=True,
        )
        self.parameters.add_parameter(
            "beta1",
            start_val=beta1,
            enabled=True,
        )
        self.parameters.add_parameter(
            "beta2",
            start_val=beta2,
            enabled=True,
        )

    def _to_xml_element(self):
        return super()._to_xml_element(name="Mishin-Cu-rho")


class MishinCuF(SpecialFunction):
    def __init__(
        self,
        identifier=None,
        F0=None,
        F2=None,
        q1=None,
        q2=None,
        q3=None,
        q4=None,
        Q1=None,
        Q2=None,
        species=["*"],
    ):
        super().__init__(identifier, species=species, is_screening_function=False)
        self.parameters.add_parameter(
            "F0",
            start_val=F0,
            enabled=True,
        )
        self.parameters.add_parameter(
            "F2",
            start_val=F2,
            enabled=True,
        )
        self.parameters.add_parameter(
            "q1",
            start_val=q1,
            enabled=True,
        )
        self.parameters.add_parameter(
            "q2",
            start_val=q2,
            enabled=True,
        )
        self.parameters.add_parameter(
            "q3",
            start_val=q3,
            enabled=True,
        )
        self.parameters.add_parameter(
            "q4",
            start_val=q4,
            enabled=True,
        )
        self.parameters.add_parameter(
            "Q1",
            start_val=Q1,
            enabled=True,
        )
        self.parameters.add_parameter(
            "Q2",
            start_val=Q2,
            enabled=True,
        )

    def _to_xml_element(self):
        return super()._to_xml_element(name="Mishin-Cu-F")


class ExtendedMishinCuF(SpecialFunction):
    def __init__(
        self,
        identifier=None,
        F0=None,
        F2=None,
        f3=None,
        f4=None,
        f5=None,
        f6=None,
        a3=None,
        a4=None,
        a5=None,
        a6=None,
        d3=None,
        d4=None,
        d5=None,
        species=["*"],
    ):
        super().__init__(identifier, species=species, is_screening_function=False)
        self.parameters.add_parameter(
            "F0",
            start_val=F0,
            enabled=True,
        )
        self.parameters.add_parameter(
            "F2",
            start_val=F2,
            enabled=True,
        )
        self.parameters.add_parameter(
            "f3",
            start_val=f3,
            enabled=True,
        )
        self.parameters.add_parameter(
            "f4",
            start_val=f4,
            enabled=True,
        )
        self.parameters.add_parameter(
            "f5",
            start_val=f5,
            enabled=True,
        )
        self.parameters.add_parameter(
            "f6",
            start_val=f6,
            enabled=True,
        )
        self.parameters.add_parameter(
            "a3",
            start_val=a3,
            enabled=True,
        )
        self.parameters.add_parameter(
            "a4",
            start_val=a4,
            enabled=True,
        )
        self.parameters.add_parameter(
            "a5",
            start_val=a5,
            enabled=True,
        )
        self.parameters.add_parameter(
            "a6",
            start_val=a6,
            enabled=True,
        )
        self.parameters.add_parameter(
            "d3",
            start_val=d3,
            enabled=True,
        )
        self.parameters.add_parameter(
            "d4",
            start_val=d4,
            enabled=True,
        )
        self.parameters.add_parameter(
            "d5",
            start_val=d5,
            enabled=True,
        )

    def _to_xml_element(self):
        return super()._to_xml_element(name="Extended-Mishin-Cu-F")


class UserFunction(DataContainer, BaseFunctionMixin):
    """
    Analytic functions that are not implemented in atomicrex
    can be provided as user functions.
    All parameters defined in the function should be added using the
    UserFunction.parameters.add_parameter() method.
    """

    def __init__(
        self,
        identifier=None,
        input_variable=None,
        species=["*", "*"],
        is_screening_function=False,
        cutoff=None,
    ):
        super().__init__(table_name=f"user_func_{identifier}")
        self.input_variable = input_variable
        self.identifier = identifier
        self.species = species
        self.parameters = FunctionParameterList()
        self.expression = None
        self.derivative = None
        self.is_screening_function = is_screening_function
        self.cutoff = cutoff
        if not is_screening_function:
            self.screening = None

    def _to_xml_element(self):
        if self.is_screening_function:
            screening = ET.Element("screening")
            root = ET.SubElement(screening, "user-function")
        else:
            root = ET.Element("user-function")
        root.set("id", f"{self.identifier}")
        input_var = ET.SubElement(root, "input-var")
        input_var.text = f"{self.input_variable}"
        expression = ET.SubElement(root, "expression")
        expression.text = f"{self.expression}"
        derivative = ET.SubElement(root, "derivative")
        derivative.text = f"{self.derivative}"

        for param in self.parameters.values():
            p = ET.SubElement(root, "param")
            p.set("name", f"{param.param}")
            p.text = f"{param.start_val:.6g}"  # 6g formatting because atomicrex output is limited to 6 significant digits, prevents some errors

        root.append(self.parameters.fit_dofs_to_xml_element())

        if self.cutoff is not None:
            cutoff = ET.SubElement(root, "cutoff")
            cutoff.text = f"{self.cutoff}"

        if not self.is_screening_function:
            if self.screening is not None:
                root.append(self.screening._to_xml_element())
            return root
        else:
            return screening

    def _parse_final_parameter(self, leftover, value):
        param = leftover[0].rstrip(":")
        self.parameters[param].final_value = value


class FunctionParameter(DataContainer):
    """
    Function parameter. For detailed information
    about the attributes see the atomicrex documentation.
    Objects should only be created using the add_parameter method
    of the FunctionParameterList class.
    """

    def __init__(
        self,
        param=None,
        start_val=None,
        enabled=True,
        reset=False,
        min_val=None,
        max_val=None,
        fitable=True,
        tag=None,
    ):
        self.param = param
        self.start_val = start_val
        self.enabled = enabled
        self.reset = reset
        self.min_val = min_val
        self.max_val = max_val
        self.tag = tag
        self.fitable = fitable
        self.final_value = None

    def _to_xml_element(self):
        root = ET.Element(f"{self.param}")
        if self.enabled:
            root.set("enabled", "true")
        else:
            root.set("enabled", "false")

        if self.reset:
            root.set("reset", "true")
        else:
            root.set("reset", "false")

        if self.min_val is not None:
            root.set("min", f"{self.min_val:.6g}")
        if self.max_val is not None:
            root.set("max", f"{self.max_val:.6g}")
        if self.tag is not None:
            root.set("tag", f"{self.tag}")
        return root

    def copy_final_to_start_value(self, filter_func=None):
        """
        Copies the final value to start_val.

        Raises:
            ValueError: Raises if fitting of the parameter is enabled,
                        but the final value is None. This should only be the case
                        if the job aborted or was not run yet.
        """
        if filter_func is not None:
            if not filter_func(self):
                return

        if self.final_value is None:
            if self.enabled:
                raise ValueError(
                    f"Fitting is enabled for {self.param}, but final value is None."
                )
        else:
            self.start_val = copy.copy(self.final_value)

    def set_max_val(self, constant=None, factor=None, filter_func=None):
        if filter_func is not None:
            if not filter_func(self):
                return

        self.max_val = constant
        if factor is not None:
            self.max_val = abs(self.start_val) * factor

    def set_min_val(self, constant=None, factor=None, filter_func=None):
        if filter_func is not None:
            if not filter_func(self):
                return

        self.min_val = constant
        if factor is not None:
            self.min_val = -abs(self.start_val) * factor

    def lock(self, filter_func=None):
        if filter_func is not None:
            if not filter_func(self):
                return
        self.enabled = False


class FunctionParameterList(DataContainer):
    def __init__(self):
        super().__init__(table_name="FunctionParameterList")

    def add_parameter(
        self,
        param,
        start_val,
        enabled=True,
        reset=False,
        min_val=None,
        max_val=None,
        tag=None,
        fitable=True,
    ):
        """
        Add a function parameter named param to a function.
        This needs to be done manually for user functions and
        not for special functions.

        Args:
            param (str): Name of the parameter. Must exactly match the name in the function expression.
            start_val (float): Starting value of the parameter
            enabled (bool, optional): Determines if the paremeter is varied during fitting. Defaults to True.
            reset (bool, optional): Determine if the parameter should be reset every iteration
            Can help with global optimization. Defaults to False.
            min_val (float, optional): Highly recommended for global optimization. Defaults to None.
            max_val (float, optional): Highly recommended for global optimization. Defaults to None.
            tag (str, optional): [description]. Only necessary for ABOP potentials .Defaults to None.
            fitable (bool, optional): [description]. Changing could cause bugs. Defaults to True.
        """
        self[param] = FunctionParameter(
            param,
            start_val,
            enabled=enabled,
            reset=reset,
            min_val=min_val,
            max_val=max_val,
            tag=tag,
            fitable=fitable,
        )

    def fit_dofs_to_xml_element(self):
        """Internal function
        Returns fit dofs as atomicrex xml element.
        """
        fit_dof = ET.Element("fit-dof")
        for param in self.values():
            if param.fitable:
                fit_dof.append(param._to_xml_element())
        return fit_dof


class PolyCoeff(FunctionParameter):
    """
    Function parameter, but for polynomial interpolation.
    """

    def __init__(
        self,
        n: int = None,
        start_val: float = None,
        enabled=True,
        reset=False,
        min_val=None,
        max_val=None,
    ):
        super().__init__(
            param="coeff",
            start_val=start_val,
            enabled=enabled,
            reset=reset,
            min_val=min_val,
            max_val=max_val,
            fitable=True,
            tag=None,
        )
        self.n = n

    def _to_xml_element(self):
        root = super()._to_xml_element()
        root.set("value", f"{self.start_val:.6g}")
        root.set("n", f"{self.n:.6g}")
        return root


class PolyCoeffList(DataContainer):
    def __init__(self):
        super().__init__(table_name="PolyCoeffList")

    def add_coeff(
        self, n, start_val, enabled=True, reset=False, min_val=None, max_val=None
    ):
        """
        Add a term in the form of a*x^n.

        Args:
            n (int): Order n of the coefficient
            start_val (float): Starting value of a.
            enabled (bool, optional): Determines if it should be fitted. Defaults to True.
            reset (bool, optional): Determines if it should be reset after each iteration. Defaults to False.
            min_val (float, optional): Highly recommended for global optimization. Defaults to None.
            max_val (float, optional): Highly recommended for global optimization. Defaults to None.
        """
        self[f"coeff_{n}"] = PolyCoeff(
            n,
            start_val,
            enabled,
            reset,
            min_val,
            max_val,
        )

    def _to_xml_element(self):
        coefficients = ET.Element("coefficients")
        for coeff in self.values():
            coefficients.append(coeff._to_xml_element())
        return coefficients


class Node(FunctionParameter):
    """
    Function parameter, but for spline interpolation.
    """

    def __init__(
        self,
        x=None,
        start_val=None,
        enabled=True,
        reset=False,
        min_val=None,
        max_val=None,
    ):
        super().__init__(
            param="node",
            start_val=start_val,
            enabled=enabled,
            reset=reset,
            min_val=min_val,
            max_val=max_val,
            fitable=True,
            tag=None,
        )
        self.x = x

    def _to_xml_element(self):
        node = super()._to_xml_element()
        node.set("x", f"{self.x:.6g}")
        node.set("y", f"{self.start_val:.6g}")
        return node


class NodeList(DataContainer):
    def __init__(self):
        super().__init__(table_name="NodeList")

    def add_node(
        self, x, start_val, enabled=True, reset=False, min_val=None, max_val=None
    ):
        """
        Add a node to the spline interpolation function.

        Args:
            x (float): x coordinate of the node. Does not change during fitting.
            start_val (float): Initial y coordinate of the node.
            enabled (bool, optional): Determines if y is changed during fitting. Defaults to True.
            reset (bool, optional): Determines if y should be reset every iteration. Defaults to False.
            min_val (float, optional): Highly recommended for global optimization. Defaults to None.
            max_val (float, optional): Highly recommended for global optimization. Defaults to None.
        """
        x = float(x)
        # atomicrex rounds output to 6 digits, so this is done here to prevent issues when reading the output.
        key = f"node_{x:.6g}"
        self[key] = Node(
            x=x,
            start_val=start_val,
            enabled=enabled,
            reset=reset,
            min_val=min_val,
            max_val=max_val,
        )
        return self[key]

    def _to_xml_element(self):
        nodes = ET.Element("nodes")
        for node in self.values():
            nodes.append(node._to_xml_element())
        return nodes

    def create_from_arrays(self, x, y, min_vals=None, max_vals=None):
        """
        Convenience function to create nodes from lists or arrays of values.
        Allows to easily start the fitting process with physically motivated values
        or values taken from previous potentials.
        Creates len(x) nodes at position x with starting values y.
        All given arrays must have the same length.

        Args:
            x (list or array): x values of the nodes
            y (list or array): corresponding y (starting) values
            min_vals ([type], optional): Highly recommended for global optimization. Defaults to None.
            max_vals ([type], optional): Highly recommended for global optimization. Defaults to None.
        """
        for i in range(len(x)):
            node = self.add_node(x[i], y[i])
            if min_vals is not None:
                node.min_val = min_vals[i]
            if max_vals is not None:
                node.max_val = max_vals[i]


def plot(func, x=np.linspace(0.01, 7.0, 351)):
    y = func(x)
    fig, ax = plt.subplots(figsize=(10, 7))
    ax.plot(x, y)
    # These defaults should be fine for most potentials
    ax.set(xlim=[0.0, 7.0], ylim=[-3.0, 3.0], xlabel="r [$\AA$]", ylabel="func(r)")
    return fig, ax



# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department

"""
Job class to run a reference jobs on all structures in a given container via an interactive job.
"""

__author__ = "Marvin Poul"
__copyright__ = "Copyright 2021, Max-Planck-Institut für Eisenforschung GmbH " \
                "- Computational Materials Design (CM) Department"
__version__ = "0.1"
__maintainer__ = "Marvin Poul"
__email__ = "poul@mpie.de"
__status__ = "development"
__date__ = "Jun 14, 2021"


from pyiron_atomistics.atomistics.structure.structurestorage import StructureStorage
from pyiron_base import GenericMaster, DataContainer

class StructureMasterInt(GenericMaster):
    """
    Runs given structures with given reference job.

    This example shows how to run a simple energy vs distance calculation for a dimer.

    >>> from pyiron_atomistics import Project
    >>> from pyiron_contrib.atomistics.atomistics.job.structurestorage import StructureStorage
    >>> pr = Project("container")
    >>> container = StructureStorage()
    >>> for d in [0.7, 0.8, 0.9, 1, 1.1, 1.2]:
    ...     container.add_structure(pr.create.structure.atoms(["Fe", "Fe"], positions=[[0,0,0], [0,0,d]], cell=[10,10,10])
    >>> ref = pr.create.job.Lammps("ref")
    >>> ref.structure = container.get_structure()
    >>> ref.calc_static()
    >>> master = pr.create.job.StructureContainerInteractive("master")
    >>> master.container = container
    >>> master.ref_job = ref
    >>> master.run()
    """

    def __init__(self, project=None, job_name=None):
        super().__init__(project=project, job_name=job_name)
        self.input = DataContainer(table_name="parameters")
        self.input.container = None

    @property
    def container(self):
        return self.input.container

    @container.setter
    def container(self, container):
        self.input.container = container

    def add_structure(self, structure):
        """
        Add a structure to the StructureStorage to calculate.  If no container set yet, create an empty one.

        Args:
            structure (Atoms): structure to add to calculation
        """
        if self.input.container is None:
            self.input.container = StructureStorage()
        self.input.container.add_structure(structure)

    def validate_ready_to_run(self):
        if self.input.container is None:
            raise ValueError("No structure container set!")
        if self.ref_job is None:
            raise ValueError("No reference job set!")
        self.ref_job.validate_ready_to_run()
        copy = self.ref_job.copy_to(
                project=self.project_hdf5,
                new_job_name=f"{self.name}_calculator",
                new_database_entry=True
        )
        self.append(copy)


    def run_static(self):
        self.status.running = True

        j = self.pop()

        j.interactive_open()
        j.interactive_enforce_structure_reset = True
        for structure in self.container.iter_structures():
            j.structure = structure
            j.run()
        j.interactive_close()

        self.status.collect = True
        self.run()

    def collect_output(self):
        self.to_hdf()

    def write_input(self):
        pass

    def to_hdf(self, hdf=None, group_name=None):
        super().to_hdf(hdf=hdf, group_name=group_name)
        self.input.to_hdf(hdf=self.project_hdf5)

    def from_hdf(self, hdf=None, group_name=None):
        super().from_hdf(hdf=hdf, group_name=group_name)
        self.input.from_hdf(hdf=self.project_hdf5)



# coding: utf-8
# Copyright (c) Georg-August-Universität Göttingen - Behler Group
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""Class Overrides for all `runnerase` storage classes.

The ASE calculator for RuNNer, provided by the Python package `runnerase`,
stores data like the values of symmetry functions or the weights of atomic
neural networks in custom classes. However, these classes generally do not know
how to write themselves to HDF5 storage format. Therefore, these classes are
extended in this module with additional functions, typically `to_hdf(...)` and
`from_hdf(...)`.

Attributes:
     HDFSymmetryFunctionValues (FlattenedStorage): Storage container for
        symmetry function values.
    RunneraseHDFMixin (HasHDF): Abstract mixin for all classes that store
        RuNNer results to HDF.
    HDFSymmetryFunctionSet (SymmetryFunctionSet, RunneraseHDFMixin): Storage
        container for a set of symmetry functions.
    HDFSplitTrainTest (RunnerSplitTrainTest, RunneraseHDFMixin): Storage
        container for the splitting between training and testing dataset.
    HDFFitResults (RunnerFitResults, RunneraseHDFMixin): Storage container
        for the results of a RuNNer fit.
    HDFWeights (RunnerWeights, RunneraseHDFMixin): Storage container for the
        weights of atomic neural networks.
    HDFScaling (RunnerScaling, RunneraseHDFMixin): Storage container for the
        symmetry function scaling data.

.. _RuNNer online documentation:
   https://theochem.gitlab.io/runner
"""

from typing import Optional, Union
from abc import abstractmethod

import numpy as np

from runnerase.symmetryfunctions import SymmetryFunctionSet
from runnerase.storageclasses import (RunnerSymmetryFunctionValues,
                                      RunnerStructureSymmetryFunctionValues,
                                      RunnerSplitTrainTest,
                                      RunnerFitResults,
                                      RunnerWeights, RunnerScaling)
from pyiron_base import ProjectHDFio
from pyiron_base import HasHDF
from pyiron_base import FlattenedStorage

from .utils import pad, unpad


class HDFSymmetryFunctionValues(FlattenedStorage):
    """Extend runnerase RunnerSymmetryFunctionValues with HDF5 compatibility."""

    __hdf_version__ = '0.3.0'

    def from_runnerase(
        self,
        runnerase_sfvalues: RunnerSymmetryFunctionValues
    ) -> None:
        """Fill `self` with information of the corresponding `runnerase` object.

        `runnerase` stores the symmetry function values of each structure in
        a separate `RunnerStructureSymmetryFunctionValues` object. However,
        it is very inefficient to read and write all of this information
        separately into HDF5 storage. Therefore, we use a `FlattenedStorage`
        where all the symmetry function values of all structures are stored
        in one large flat array.

        Matters are complicated further because `FlattenedStorage` only accepts
        data chunks (read here: structures) where all different arrays of
        information have the same length. Unfortunately, one can have different
        numbers of symmetry functions for each element and varying numbers of
        elements for each structure in a training dataset.

        The solution here is to `pad` the sfvalues of each group of elements in
        one structure with rows filled with `np.NaN`. This way, all arrays for
        one structure have length `number_of_atoms` and can be efficiently
        stored and retrieved even for large  datasets.

        Parameters
        ----------
        runnerase_sfvalues : RunnerSymmetryFunctionValues
            A `runnerase` class object containing the symmetry function values
            for a whole dataset.
        """
        for structure_sfvalues in runnerase_sfvalues.data:

            # Preprocess the symmetry function value arrays.
            sfvalues_arrays = {}
            for element, sfvalues in structure_sfvalues.data.items():
                name = f'sfvalues_{element}'
                shape = (sfvalues.shape[1],)
                # Add arrays for storing the symmetry function values of all
                # atoms of the same `element`, unless they are already present.
                if name not in self.list_arrays():
                    self.add_array(name, shape=shape, dtype=np.float64,
                                   per='element', fill=np.NaN)

                # Pad the sfvalues of the current `element` with `np.NaN` rows,
                # so that the total length is equal to the number of atoms in
                # the structure.
                sfvalues_padded = pad(sfvalues, len(structure_sfvalues))

                sfvalues_arrays[name] = sfvalues_padded

            self.add_chunk(
                len(structure_sfvalues),
                energy_total=structure_sfvalues.energy_total,
                energy_short=structure_sfvalues.energy_short,
                energy_elec=structure_sfvalues.energy_elec,
                charge=structure_sfvalues.charge,
                **sfvalues_arrays
            )

    def to_runnerase(self) -> RunnerSymmetryFunctionValues:
        """Create the corresponding `runnerase` object from `self`.

        Returns
        ----------
        runnerase_sfvalues : RunnerSymmetryFunctionValues
            A `runnerase` class object containing the symmetry function values
            for a whole dataset.
        """
        runnerase_sfvalues = RunnerSymmetryFunctionValues()

        for chunk_idx in range(len(self)):
            # Create a new object for storing the sfvalues of one structure.
            struct_sfvalues = RunnerStructureSymmetryFunctionValues()

            # Fill the object with per-chunk properties.
            struct_sfvalues.energy_total = self['energy_total', chunk_idx]
            struct_sfvalues.energy_short = self['energy_short', chunk_idx]
            struct_sfvalues.energy_elec = self['energy_elec', chunk_idx]
            struct_sfvalues.charge = self['charge', chunk_idx]

            # Read the symmetry function values.
            for arrayname in self.list_arrays():
                if arrayname.startswith('sfvalues'):
                    element = arrayname.split('_')[1]
                    sfvalues = self[f'sfvalues_{element}', chunk_idx]
                    struct_sfvalues.data[element] = unpad(sfvalues)

            # Append the structure to the container object
            # `RunnerSymmetryFunctionValues`.
            runnerase_sfvalues.data.append(struct_sfvalues)

        return runnerase_sfvalues


class RunneraseHDFMixin(HasHDF):
    """Abstract Mixin to add HDF5 compatibility to runnerase classes."""

    __hdf_version__ = '0.3.0'

    @property
    @abstractmethod
    def runnerase_properties(self):
        """Define the class properties stored in `self.baseclass` objects."""
        ...

    @property
    @abstractmethod
    def baseclass(self):
        """Define the runnerase class which is wrapped by this HDF class."""
        ...

    def from_runnerase(
        self,
        runnerase_class: Union[RunnerSplitTrainTest, RunnerWeights,
                               RunnerScaling, RunnerFitResults,
                               RunnerSymmetryFunctionValues]
    ) -> None:
        """Fill `self` with information of the corresponding `runnerase` object.

        Args:
            runnerase_class (runnerase storage class): The runnerase class whose
                information will be wrapped.
        """
        for prop in self.runnerase_properties:
            self.__dict__[prop] = runnerase_class.__dict__[prop]

    def to_runnerase(self) -> Union[RunnerSplitTrainTest, RunnerWeights,
                                    RunnerScaling, RunnerFitResults,
                                    RunnerSymmetryFunctionValues]:
        """Create the corresponding `runnerase` object from `self`.

        Returns:
            runnerase_class (runnerase storage class): The runnerase class whose
                information was wrapped.
        """
        runnerase_class = self.baseclass()

        for prop in self.runnerase_properties:
            runnerase_class.__dict__[prop] = self.__dict__[prop]

        return runnerase_class

    def _to_hdf(self, hdf: ProjectHDFio) -> None:
        """Write `self` to HDF5 storage.

        Args:
            hdf (ProjectHDFio): The HDF file where `self` will be stored.
        """
        for prop in self.runnerase_properties:
            hdf[f'{prop}'] = self.__dict__[prop]

    def _from_hdf(
        self,
        hdf: ProjectHDFio,
        version: Optional[str] = None
    ) -> Union[RunnerSplitTrainTest, RunnerWeights, RunnerScaling,
               RunnerFitResults, RunnerSymmetryFunctionValues]:
        """Read `self` from HDF5 storage.

        Args:
            hdf (ProjectHDFio): The HDF file where `self` will be stored.
            version (str): The HDF version of the storage file.
        """
        if version != self.__hdf_version__:
            raise RuntimeError('Invalid HDF5 version found while reading '
                               + self.__class__.__name__)

        # Open HDF file at the right group with a context manager.
        for node in hdf.list_nodes():
            for prop in self.runnerase_properties:
                if prop in node:
                    self.__dict__[prop] = hdf[node]

        return self

    def _get_hdf_group_name(self):
        """Get the name of the group where this object is stored in HDF."""
        return self.__class__.__name__


class HDFSymmetryFunctionSet(SymmetryFunctionSet, RunneraseHDFMixin):
    """Extend runnerase SymmetryFunctionSet with HDF5 compatibility."""

    __hdf_version__ = '0.3.0'

    @property
    def runnerase_properties(self):
        """Show class properties stored in `SymmetryFunctionSet` objects."""
        return ['_sets', '_symmetryfunctions', 'min_distances']

    @property
    def baseclass(self):
        """Define the base class which is wrapped by this HDF class."""
        return SymmetryFunctionSet

    def _to_hdf(
        self,
        hdf: ProjectHDFio
    ) -> None:
        """Write `self` to HDF5 storage.

        `runnerase`s SymmetryFunctionSet has the convenient property that
        all symmetry functions can be written to and read from a list
        representation. Therefore, they are also stored as lists in HDF format.

        Args:
            hdf (ProjectHDFio): The HDF file where `self` will be stored.
        """
        for idx, sfset in enumerate(self.sets):
            hdfset = HDFSymmetryFunctionSet()
            hdfset.from_runnerase(sfset)
            hdfset.to_hdf(hdf=hdf, group_name=f'set__index_{idx}')

        symmetryfunctions = [sf.to_list() for sf in self.symmetryfunctions]
        hdf['symmetryfunctions__index_0'] = symmetryfunctions

    def _from_hdf(
        self,
        hdf: ProjectHDFio,
        version: Optional[str] = None
    ) -> 'HDFSplitTrainTest':
        """Read `self` from HDF5 storage.

        Args:
            hdf (ProjectHDFio): The HDF file where `self` will be stored.
            version (str): The HDF version of the storage file.
        """
        if version != self.__hdf_version__:
            raise RuntimeError('Invalid HDF5 version found while reading '
                               + self.__class__.__name__)

        # Reload symmetry function sets.
        for group in hdf.list_groups():
            if group.startswith('set'):
                new_set = hdf.__getitem__(group).to_object()
                self.append(new_set)

        for node in hdf.list_nodes():
            # Reload symmetry functions.
            if node.startswith('symmetryfunctions'):
                sflist = hdf.__getitem__(node)
                self.from_list(sflist)

        return self


class HDFSplitTrainTest(RunnerSplitTrainTest, RunneraseHDFMixin):
    """Mix HDF5 compatibility into RunnerSplitTrainTest`."""

    @property
    def runnerase_properties(self):
        """Show class properties stored in `RunnerSplitTrainTest` objects."""
        return ['train', 'test']

    @property
    def baseclass(self):
        """Define the base class which is wrapped by this HDF class."""
        return RunnerSplitTrainTest


class HDFFitResults(RunnerFitResults, RunneraseHDFMixin):
    """Mix HDF5 compatibility into RunnerFitResults`."""

    @property
    def runnerase_properties(self):
        """Show class properties stored in `RunnerFitResults` objects."""
        return ['epochs', 'rmse_energy', 'rmse_forces', 'rmse_charge',
                'opt_rmse_epoch', 'units']

    @property
    def baseclass(self):
        """Define the base class which is wrapped by this HDF class."""
        return RunnerFitResults


class HDFWeights(RunnerWeights, RunneraseHDFMixin):
    """Mix HDF5 compatibility into RunnerWeights`."""

    @property
    def runnerase_properties(self):
        """Show class properties stored in `RunnerWeights` objects."""
        return ['data']

    @property
    def baseclass(self):
        """Define the base class which is wrapped by this HDF class."""
        return RunnerWeights


class HDFScaling(RunnerScaling, RunneraseHDFMixin):
    """Mix HDF5 compatibility into RunnerScaling`."""

    @property
    def runnerase_properties(self):
        """Show class properties stored in `RunnerScaling` objects."""
        return ['data', 'target_min', 'target_max']

    @property
    def baseclass(self):
        """Define the base class which is wrapped by this HDF class."""
        return RunnerScaling



from abc import ABC, abstractmethod
from json import dumps


class HasToDict(ABC):
    @abstractmethod
    def to_dict(self):
        pass

    def _repr_json_(self):
        return self.to_dict()

    @property
    def info(self):
        print(dumps(self.to_dict(), indent=2))



import os
from pyiron_base import DataContainer, state
from pyiron_atomistics.atomistics.structure.atoms import pyiron_to_ase, ase_to_pyiron
from pyiron_atomistics.atomistics.job.atomistic import AtomisticGenericJob
from pycp2k import CP2K


pyiron_dict = {
    "global": {
        "run_type": "ENERGY_FORCE"
    },
    "force_eval": {
        "method": "Quickstep",
        "print_forces_section": "ON"
    },
    "dft": {
        "qs": {
            "eps": 1.0E-10
        },
        "mgrid": {
            "ngrids": 4,
            "cutoff": 300,
            "rel_cutoff": 60
        },
        "xc": {
            "functional": "PADE"
        }
    },
    "scf": {
        "scf_guess": "ATOMIC",
        "eps": 1.0E-7,
        "max": 300,
        "diagonalization": {
            "algorthim": "STANDARD"
        },
        "mixing": {
            "method": "BROYDEN_MIXING",
            "alpha": 0.4,
            "n_broyden": 8,
        }
    },
    "kind": {
        "basis_set": "DZVP-GTH-PADE",
        "potential": "GTH-PADE-q4"
    }
}


class Cp2kJob(AtomisticGenericJob):
    def __init__(self, project, job_name):
        super(Cp2kJob, self).__init__(project, job_name)
        self.__name__ = "cp2k"
        self.input = DataContainer(pyiron_dict, table_name="control_dict")

    def to_hdf(self, hdf=None, group_name=None):
        """
        Store the ExampleJob object in the HDF5 File

        Args:
            hdf (ProjectHDFio): HDF5 group object - optional
            group_name (str): HDF5 subgroup name - optional
        """
        super(Cp2kJob, self).to_hdf(hdf=hdf, group_name=group_name)
        self._structure_to_hdf()
        with self.project_hdf5.open("input") as hdf5_input:
            self.input.to_hdf(hdf5_input)

    def from_hdf(self, hdf=None, group_name=None):
        """
        Restore the ExampleJob object in the HDF5 File

        Args:
            hdf (ProjectHDFio): HDF5 group object - optional
            group_name (str): HDF5 subgroup name - optional
        """
        super(Cp2kJob, self).from_hdf(hdf=hdf, group_name=group_name)
        self._structure_from_hdf()
        with self.project_hdf5.open("input") as hdf5_input:
            self.input.from_hdf(hdf5_input)

    def write_input(self):
        calc = CP2K()
        calc.working_directory = self.working_directory
        calc.project_name = "pyiron"
        CP2K_INPUT = calc.CP2K_INPUT
        GLOBAL = CP2K_INPUT.GLOBAL
        FORCE_EVAL = CP2K_INPUT.FORCE_EVAL_add()  # Repeatable items have to be first created
        SUBSYS = FORCE_EVAL.SUBSYS
        DFT = FORCE_EVAL.DFT
        SCF = DFT.SCF
        GLOBAL.Run_type = self.input["global"]["run_type"]
        FORCE_EVAL.Method = self.input["force_eval"]["method"]
        FORCE_EVAL.PRINT.FORCES.Section_parameters = self.input["force_eval"]["print_forces_section"]
        DFT.Basis_set_file_name = os.path.join(state.settings.resource_paths[0], "cp2k", "potentials", "BASIS_SET")
        DFT.Potential_file_name = os.path.join(state.settings.resource_paths[0], "cp2k", "potentials", "GTH_POTENTIALS")
        DFT.QS.Eps_default = self.input["dft"]["qs"]["eps"]
        DFT.MGRID.Ngrids = self.input["dft"]["mgrid"]["ngrids"]
        DFT.MGRID.Cutoff = self.input["dft"]["mgrid"]["cutoff"]
        DFT.MGRID.Rel_cutoff = self.input["dft"]["mgrid"]["rel_cutoff"]
        DFT.XC.XC_FUNCTIONAL.Section_parameters = self.input["dft"]["xc"]["functional"]
        SCF.Scf_guess = self.input["scf"]["scf_guess"]
        SCF.Eps_scf = self.input["scf"]["eps"]
        SCF.Max_scf = self.input["scf"]["max"]
        SCF.DIAGONALIZATION.Section_parameters = "ON"
        SCF.DIAGONALIZATION.Algorithm = self.input["scf"]["diagonalization"]["algorthim"]
        SCF.MIXING.Section_parameters = "T"
        SCF.MIXING.Method = self.input["scf"]["mixing"]["method"]
        SCF.MIXING.Alpha = self.input["scf"]["mixing"]["alpha"]
        SCF.MIXING.Nbroyden = self.input["scf"]["mixing"]["n_broyden"]
        for el in set(self.structure.get_chemical_symbols()):
            KIND = SUBSYS.KIND_add(el)
        KIND.Basis_set = self.input["kind"]["basis_set"]
        KIND.Potential = self.input["kind"]["potential"]
        calc.create_cell(SUBSYS, pyiron_to_ase(self.structure))
        calc.create_coord(SUBSYS, pyiron_to_ase(self.structure))
        calc.write_input_file()

    def collect_output(self):
        pass



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

"""
Convenience class to handle fitted MTPs.
"""

from pyiron_base import DataContainer
from pyiron_contrib.atomistics.mlip.parser import potential as parse_potential

import numpy as np

__author__ = "Marvin Poul"
__copyright__ = "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - " \
                "Computational Materials Design (CM) Department"
__version__ = "1.0"
__maintainer__ = "Marvin Poul"
__email__ = "poul@mpie.de"
__status__ = "development"
__date__ = "Aug 18, 2021"

class MtpPotential:
    """
    Representation of a fitted MTP.

    Potentials may be loaded from a file when initializing

    >>> pot = MtpPotential("some/path")

    or later with :method:`.load()`

    >>> pot = MtpPotential()
    >>> pot.load("another/path")

    This class exports the fitted radial basis and the moment coefficients

    >>> pot.moment_coefficients
    array([...])
    >>> pot.radial_basis[0]
    Chebyshev([...], domain=(...), window=(...))

    """
    def __init__(self, filename=None, table_name="potential"):
        """
        Create a new potential.

        If `filename` is not given, :method:`.load()` must be called before the class is usable.

        Args:
            filename (str, optional): from where to load the potential
            table_name (str, optional): default group name when saving to HDF, default 'potential'
        """
        self.__hdf_version__ = "0.0.1"
        self._store = DataContainer(table_name=table_name)
        if filename is not None:
            self.load(filename)

    def load(self, filename):
        with open(filename) as f:
            self._store.clear()
            self._store.update(parse_potential(f.read()), wrap=True)

    def write(self, filename):
        """
        Write potential to a text file readable by mlp.

        Args:
            filename (str): file path
        """
        def format_array(a):
            return "{" + ', '.join(map(str, a)) + "}"

        with open(filename, "w") as f:
            f.write("MTP\n")
            f.write(f"version = {self._store.version}\n")
            f.write(f"potential_name = {self._store.potential_name}\n")
            f.write(f"scaling = {self._store.scaling:.15e}\n")
            f.write(f"species_count = {self._store.species_count}\n")
            f.write(f"potential_tag = {self._store.get('potential_tag', '')}\n")
            f.write(f"radial_basis_type = RB{self._store.radial.basis_type}\n")
            f.write(f"\tmin_dist = {self._store.radial.info.min_dist:.15e}\n")
            f.write(f"\tmax_dist = {self._store.radial.info.max_dist:.15e}\n")
            f.write(f"\tradial_basis_size = {self._store.radial.info.basis_size}\n")
            f.write(f"\tradial_funcs_count = {self._store.radial.info.funcs_count}\n")
            f.write(f"\tradial_coeffs\n")
            for types, funcs in self._store.radial.funcs.items():
                f.write(f"\t\t{types}\n")
                for coeffs in funcs:
                    f.write("\t\t\t" + format_array(coeffs) + "\n")
            f.write(f"alpha_moments_count = {self._store.alpha_moments_count}\n")
            f.write(f"alpha_index_basic_count = {self._store.alpha_index_basic_count}\n")
            f.write(f"alpha_index_basic = {format_array(map(format_array, self._store.alpha_index_basic))}\n")
            f.write(f"alpha_index_times_count = {self._store.alpha_index_times_count}\n")
            f.write(f"alpha_index_times = {format_array(map(format_array, self._store.alpha_index_times))}\n")
            f.write(f"alpha_scalar_moments = {self._store.alpha_scalar_moments}\n")
            f.write(f"alpha_moment_mapping = {format_array(self._store.alpha_moment_mapping)}\n")
            f.write(f"species_coeffs = {format_array(self._store.species_coeffs)}\n")
            f.write(f"moment_coeffs = {format_array(self._store.moment_coeffs)}\n")

    def _type_to_hdf(self, hdf):
        """
        Internal helper function to save type and version in hdf root

        Args:
            hdf (ProjectHDFio): HDF5 group object
        """
        hdf["NAME"] = self.__class__.__name__
        hdf["TYPE"] = str(type(self))
        hdf["HDF_VERSION"] = self.__hdf_version__

    def to_hdf(self, hdf, group_name=None):
        if group_name is not None:
            hdf = hdf.open(group_name)
        # force DataContainer to write into hdf group directly
        self._store.to_hdf(hdf=hdf, group_name=None)
        # then overwrite type information with ours, so we get reinstantiated later
        self._type_to_hdf(hdf)
        if group_name is not None:
            hdf.close()

    def from_hdf(self, hdf, group_name=None):
        self._store.from_hdf(hdf=hdf, group_name=group_name)

    @property
    def radial_basis(self):
        """
        list of numpy polynomials: radial basis functions used to compute the moment tensors
        """
        basis_type = self._store.radial.basis_type
        rmin = self._store.radial.info.min_dist
        rmax = self._store.radial.info.max_dist
        scaling = self._store.scaling
        if basis_type == 'Chebyshev':
            return {types: [scaling * np.polynomial.Chebyshev(coeffs, domain=(rmin, rmax))
                                for coeffs in funcs]
                                    for types, funcs in self._store.radial.funcs.items()}

        else:
            raise NotImplementedError(f"unknown basis type {basis_type}")

    @property
    def moment_coefficients(self):
        """
        array of floats: expansion coefficients for moment tensor contractions
        """
        return self._store.moment_coeffs



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import fnmatch
import io
import json
import os
import posixpath
from functools import lru_cache

import boto3
from botocore.client import Config

from pyiron_contrib.generic.filedata import StorageInterface
from pyiron_contrib.generic.filedata import load_file, FileDataTemplate


class S3FileData(FileDataTemplate):
    """FileData stores an instance of a data file from an S3 system, e.g. a single Image from a measurement."""

    def __init__(self, s3obj, filename=None, filetype=None):
        """FileData class to store data and associated metadata.

        Args:
            s3obj (s3object): s3object containing a file with data
            filename (str): filename associated with the data.
            filetype (str): File extension associated with the type data,
                            If provided this overwrites the assumption based on the extension of the filename.
        """
        self._s3obj = s3obj
        if filename is None:
            self.filename = s3obj.key.split("/")[-1]
        else:
            self.filename = filename
        if filetype is None:
            self.filetype = self._get_filetype_from_filename(self.filename)
        else:
            self.filetype = filetype
        self._data = None
        self._metadata = self._s3obj.metadata

    @property
    @lru_cache()
    def data(self):
        """Return the associated data."""
        if self._data is None:
            self._s3obj = self._s3obj.get()
            self._data = self._s3obj["Body"].read()
        return load_file(io.BytesIO(self._data), filetype=self.filetype)

    @property
    def metadata(self):
        return self._metadata

    def __repr__(self):
        return f"FileData containing {self.filename}."

    def _ipython_display_(self):
        result = f"File    {self.filename}   with meta data:\n"
        result += "-------------------------------------------\n"
        for key, value in self.metadata.items():
            result += f"{key}:  {value}\n"
        print(result)


class S3ioConnect:
    def __init__(self, credentials=None, bucket_name=None):
        """Establishes connection to a specific 'bucket' of a S3 type object store.

        Args:
            credentials (str/dict/boto3.resource/None):
                    if str: path to a json configuration file with login credentials for the bucket.
                    if dict: dictionary containing the login credentials.
                    if boto3.resource: This resource is used as is, requires bucket_name to be specified.
                    if None: use Credentials as obtained by boto3, requires bucket_name to be specified.
            bucket_name(str/None): Name of the bucket, overwrites name given in the config.

        The credentials needs to provide the following information:
            access_key or aws_access_key_id (str)
            secret_key or aws_secret_access_key (str)
        And may contain these additional fields:
            endpoint or endpoint_url (str)
            bucket (str)
        Each additional keyword is passed on to boto3 as is.
        """
        self.bucket_name = None
        self.s3resource = self._init_s3resource(credentials)

        self.bucket_name = bucket_name or self.bucket_name
        self._check_for_bucket_name()
        self.bucket = self.s3resource.Bucket(self.bucket_name)

    def _init_s3resource(self, credentials):
        if isinstance(credentials, boto3.resources.base.ServiceResource):
            return credentials
        elif credentials is None:
            # boto3 looks for the (missing) credentials at  ~/.aws/credentials or at environment variables:
            # AWS_ACCESS_KEY_ID  AWS_SECRET_ACCESS_KEY  etc.
            return boto3.resource("s3")
        else:
            credentials = self._get_credentials_dict(credentials)

            config = credentials.pop("config", Config(s3={"addressing_style": "path"}))

            self._normalize_key(credentials, "aws_access_key_id", "access_key")
            self._normalize_key(credentials, "aws_secret_access_key", "secret_key")
            self._normalize_key(credentials, "endpoint_url", "endpoint")

            self.bucket_name = credentials.pop("bucket", None)

            return boto3.resource("s3", config=config, **credentials)

    def _check_for_bucket_name(self):
        if self.bucket_name is None:
            raise ValueError("Bucket name needs to be provided.")

    @staticmethod
    def _normalize_key(credentials, key, alias_key):
        value = credentials.pop(alias_key, None) or credentials.pop(key, None)
        if value is not None:
            credentials[key] = value

    @staticmethod
    def _get_credentials_dict(credentials):
        if isinstance(credentials, str):
            with open(credentials) as json_file:
                credentials = json.load(json_file)
        if not isinstance(credentials, dict):
            raise TypeError(
                f"credentials is not one of the supported types but {type(credentials)}."
            )
        return credentials.copy()

    @property
    def endpoint_url(self):
        return self.s3resource.meta.client.meta.endpoint_url


class FileS3IO(StorageInterface):
    """Provides access to a specific bucket of a S3 object store similar to a regular storage system.

    Implements :class:`.HasGroups`. Groups are 'Folders' in the S3 storage, nodes are file like objects.

    Files may be uploaded to the bucket using
        upload: upload a (list of) filenames to the current group of the bucket.
        put: upload a file like object to the specified path in the bucket.

    These methods may be used to receive data
        download: download (list of) files to the specified folder.
        get: receive a S3FileData from the storage
        get_s3_object: receive a native s3.Object from the store
        get_metadata: receive only the meta data from the specified file

    Attributes:
        s3_path: absolute path (starting with '/') inside the bucket, interpreting '/' as the directory separator.
        connection_info: dict with name and endpoint of the bucket.
    """

    @property
    def _path(self):
        return self.s3_path

    @_path.setter
    def _path(self, new_path):
        self.s3_path = new_path

    def __init__(self, config=None, path="/", *, bucket_name=None):
        """
        Establishes connection to a specific 'bucket' of a S3 type object store.

        Args:
            config (str/dict/:class:`S3IO_connect`/boto3.ressource/None):
                Provides access information for the S3 type object store:
                    str: path to a json configuration file with login credentials for the bucket.
                    dict: dictionary containing the login credentials.
                    S3IO_connect: Instantiated S3IO_connect class to access the S3 system.
                    boto3.resource: Instantiated boto3.resource, requires bucket_name to be given.
                    None: Use default credentials from boto3, requires bucket_name to be given.
            path (str): Initial group in the bucket which is opened.
            bucket_name(str/None): Name of the bucket if not included/different from the config.

        The configuration needs to provide the following information:
            {
            access_key : ""
            secret_key : ""
            endpoint : ""
            bucket : ""
            }
        """
        super().__init__()
        self._path = path
        self.history = [path]
        if isinstance(config, S3ioConnect):
            self._s3io = config
        else:
            self._s3io = S3ioConnect(credentials=config, bucket_name=bucket_name)

        self._bucket = self._s3io.bucket
        self._s3_path = None
        self.s3_path = path

    @property
    def s3_path(self):
        """
        Get the path in the S3 object store starting from the root group - meaning this path starts with '/'

        Returns:
            str: S3 path
        """
        return self._s3_path

    @s3_path.setter
    def s3_path(self, path):
        """
        Set the path in the S3 object store starting from the root group

        Args:
            path (str): S3 path
        """
        if (path is None) or (path == ""):
            path = "/"
        self._s3_path = posixpath.normpath(path)
        if not posixpath.isabs(self._s3_path):
            self._s3_path = "/" + self._s3_path
        if not self._s3_path[-1] == "/":
            self._s3_path = self._s3_path + "/"

    @property
    def _bucket_path(self):
        """
        The bucket object internally does not use a '/' to indicate the root group.

        Return:
            str: Internal path in the bucket.
        """
        return self._s3_path[1:]

    @property
    def connection_info(self):
        return {
            "s3_path": self.s3_path,
            "bucket_name": self._bucket.name,
            "endpoint_url": self._s3io.endpoint_url,
        }

    @classmethod
    def from_dict(cls, connection_dict: dict):
        path = connection_dict.get('s3_path', '/')
        config = {
            'bucket': connection_dict.get('bucket_name', None),
            'endpoint':  connection_dict.get('endpoint_url', None)
        }
        cls(config, path)

    def _list_groups(self):
        """
        List directories/groups in the current group.

        Returns:
            list: list of directory names.
        """
        groups = []
        group_path_len = len(self._bucket_path.split("/")) - 1
        for obj in self._list_objects():
            rel_obj_path_spl = obj.key.split("/")[group_path_len:]
            if len(rel_obj_path_spl) > 1:
                if rel_obj_path_spl[0] not in groups:
                    groups.append(rel_obj_path_spl[0])
        return groups

    def _list_nodes(self):
        """
        List of 'files' ( string not followed by '/' ) in the current group.

        Returns:
            list: list of file names.
        """
        nodes = []
        for obj in self._bucket.objects.filter(Prefix=self._bucket_path, Delimiter="/"):
            nodes.append(obj.key.split("/")[-1])
        return nodes

    def _to_abs_bucketpath(self, path):
        """Helper function to convert a given path to an absolute path inside the S3 bucket."""
        if path is None or "":
            path = self._bucket_path
        if posixpath.isabs(path):
            path = path[1:]
        else:
            path = self._bucket_path + path
        return path

    def is_dir(self, path):
        """
        Check if given path is a directory.

        Args:
            path (str): path to check.

        Returns:
            bool: True if path is a directory.
        """
        path = self._to_abs_bucketpath(path)
        if len(path) > 1 and path[-1] != "/":
            path = path + "/"
        objs = list(self._bucket.objects.filter(Prefix=path))
        return len(objs) > 0

    def is_file(self, path):
        """
        Check if given path is a file.

        Args:
            path (str): path to check.

        Returns:
            bool: True if path is a file.
        """
        path = self._to_abs_bucketpath(path)
        objs = list(self._bucket.objects.filter(Prefix=path))
        for obj in objs:
            if obj.key == path:
                return True
        return False

    def open(self, group):
        """
        Opens the provided group (create group if not yet present).

        Args:
            group (str): group to open/create.
        """
        new = self.copy()
        new.s3_path = self.s3_path + group
        new.history.append(new.s3_path)
        return new

    def copy(self):
        """
        Copy the Python object which links to the S3 object store.

        Returns:
            FileS3IO: New FileS3io object pointing to the same S3 object store
        """
        new = FileS3IO(config=self._s3io, path=self.s3_path)
        return new

    def close(self):
        """Close current group and open previous group."""
        if len(self.history) > 1:
            del self.history[-1]
        elif len(self.history) == 1:
            self.history[0] = "/"
        self._s3_path = self.history[-1]

    def validate_metadata(self, metadata, raise_error=True):
        if metadata is None:
            return {}
        elif not isinstance(metadata, dict) and raise_error:
            raise ValueError(
                f"Meta data needs to be a dict containing only strings but got type {type(metadata)}."
            )
        elif not isinstance(metadata, dict):
            return None
        elif not all([isinstance(i, str) for i in metadata]):
            raise ValueError(f"Meta data only allows keywords of type 'str'.")

    def upload_file(self, files, metadata=None, filenames=None):
        """
        Uploads files into the current group of the S3 object store.

        Arguments:
            files (list/str) : file / List of files to upload
            metadata (dictionary): metadata of the files (Not nested, only "str" type)
            filenames (list/str/None): Names the files should get
        """

        if isinstance(files, str):
            files = [files]

        if filenames is None:
            filenames = [os.path.basename(file) for file in files]
        elif isinstance(filenames, str):
            filenames = [filenames]

        if isinstance(metadata, list) and len(metadata) != len(files):
            raise ValueError("length of files and meta data have to match!")
        elif isinstance(metadata, list):
            metadata = [self.validate_metadata(data) for data in metadata]
        else:
            tmp_metadata = self.validate_metadata(metadata)
            metadata = [tmp_metadata for _ in filenames]

        if len(filenames) != len(files):
            raise ValueError("length of files and of filenames have to match!")

        for file, filename, _metadata in zip(files, filenames, metadata):

            self._bucket.upload_file(
                file, self._bucket_path + filename, {"Metadata": _metadata}
            )

    def download_file(self, files, targetpath="."):
        """
        Download files from current group to local file system (current directory is default)

        Arguments:
            files (list/str): List of filenames in the S3 object store.
            targetpath (str): Path in the local data system, to which the files should be downloaded.
        """
        if not os.path.exists(targetpath):
            os.mkdir(targetpath)
        if isinstance(files, str):
            files = [files]

        for f in files:
            filepath = os.path.join(targetpath, f.split("/")[-1])
            print(filepath)
            self._bucket.download_file(self._bucket_path + f, filepath)

    def get_metadata(self, file):
        """
        Returns the metadata of a file.

        Args:
            file (str): path to a file of the bucket.
        Returns:
             dict: metadata field associated with the file.
        """
        file = self._to_abs_bucketpath(file)
        return self._bucket.Object(file).metadata

    def get(self, file):
        """
        Returns an :class:`S3FileData` object which contains the file.

        Args:
            file (str): path to a file of the bucket.
        Returns:
        """
        file = self._to_abs_bucketpath(file)
        s3object = S3FileData(s3obj=self._bucket.Object(file))
        return s3object

    def get_s3_object(self, file):
        """
        Returns a s3.Object containing the requested file.

        Args:
            file(str): a path like string.
        Returns:
             Object containing a file.
        """
        file = self._to_abs_bucketpath(file)
        return self._bucket.Object(file).get()

    def put(self, data_obj, filename=None, path=None, metadata=None):
        """
        Upload a data_obj to the current group/ the provided path.

        Args:
            data_obj( <class 'bytes'>, <class 'bytearray'>, file-like object): data object to upload the data from.
            filename(str/None): Name under which the data_obj will be stored.
            path(str/None): path to upload the data to, None means currently opened group.
            metadata(dict/None): metadata to be used (has to be a dictionary of type {"string": "string", }).
                  Provided metadata overwrites the one possibly present in the data object.
        """
        if self.is_dir(path):
            path = self._to_abs_bucketpath(path)
        else:
            raise ValueError("No valid path specified!")
        if len(path) > 0 and path[-1] != "/":
            path = path + "/"

        if filename is not None:
            filename_ = filename
        elif hasattr(data_obj, "name"):
            filename_ = os.path.basename(data_obj.name)
        else:
            raise ValueError("No filename given.")
        path = path + filename_

        if metadata is None:
            metadata = {}
        self._bucket.put_object(Key=path, Body=data_obj, Metadata=metadata)

    def _list_objects(self):
        return self._bucket.objects.filter(Prefix=self._bucket_path)

    def print_fileinfos(self):
        """
        Prints the filename, last modified date and size for all files in the current group,
        recursively including sub groups.
        """
        for obj in self._bucket.objects.filter(Prefix=self._bucket_path):
            print(f"/{obj.key} {obj.last_modified} {obj.size} bytes")

    def _list_all_files_of_bucket(self):
        return list(self._bucket.objects.all())

    def glob(self, path):
        """
        Return a list of paths matching a pathname pattern.
        The pattern may contain simple shell-style wildcards a la fnmatch.

        Args:
            path(str): a path like string which may contain shell-style wildcards.

        Return:
            list: List of file names (str) matching the provided path pattern.
        """
        path = self._to_abs_bucketpath(path)
        l = []
        for obj in self._bucket.objects.filter(Prefix=self._bucket_path):
            if fnmatch.fnmatchcase(obj.key, path):
                l.append(obj.key)
        return l

    @staticmethod
    def print_file_info(filelist):
        """
        Prints filename, last_modified, and size of each file in the provided list of file objects.

        Args:
            filelist (list): List containing objects from a bucket.
        """
        for obj in filelist:
            print(f"/{obj.key} {obj.last_modified} {obj.size} bytes")

    def remove_file(self, file):
        """
        Deletes the object associated with a file.

        Args:
            file (str/None): path like string to the file to be removed.
        """
        if not self.is_file(file):
            raise ValueError(f"{file} is not a file.")
        file = self._to_abs_bucketpath(file)
        self._bucket.Object(file).delete()
        # self._remove_object(prefix=file, debug=debug)

    def remove_group(self, path=None, debug=False):
        """
        Deletes the current group with all it's content recursively.

        Args:
            path (str/None): group to be removed recursively.
            debug(bool): If True, additional information is printed.
        """
        if path is None:
            path = self._s3_path
        if not self.is_dir(path):
            raise ValueError(f"{path} is not a group.")
        path = self._to_abs_bucketpath(path)
        self._remove_object(prefix=path, debug=debug)

    def _remove_object(self, prefix, debug=False):
        """
        Deletes all objects matching the provided prefix.

        Args:
            prefix(str): All objects with this prefix will be removed.
            debug(bool): If True, additional information is printed.
        """
        if debug:
            print(
                f"\nDeleting all objects with sample prefix {self._bucket.name}/{prefix}."
            )
        delete_responses = self._bucket.objects.filter(Prefix=prefix).delete()
        if debug:
            for delete_response in delete_responses:
                for deleted in delete_response["Deleted"]:
                    print(f'\t Deleted: {deleted["Key"]}')

    def __enter__(self):
        """Compatibility function for the with statement."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Compatibility function for the with statement."""
        self.close()

    def __repr__(self):
        """
        Human readable string representation.

        Return:
            str: list all nodes and groups as string.
        """
        return str(self.list_all())

    def __getitem__(self, item):
        """
        Get/ read (meta) data from the S3 object store

        Args:
            item (str, slice): path to the data or key of the data object

        Returns:
            dict/s3.Object:  meta data or data object
        """
        if isinstance(item, slice):
            raise NotImplementedError("Implement if needed, e.g. for [:]")
        else:
            item_lst = item.split("/")
            if len(item_lst) == 1 and item_lst[0] != "..":
                if item == "":
                    return self
                if item in self.list_nodes():
                    return self.get(item)
                if item in self.list_groups():
                    return self.open(item)
                raise ValueError(f"Unknown item: {item}")
            else:
                item_abs_lst = (
                    os.path.normpath(os.path.join(self.s3_path, item))
                    .replace("\\", "/")
                    .split("/")
                )
                s3_object = self.copy()
                s3_object.s3_path = "/".join(item_abs_lst[:-1])
                return s3_object[item_abs_lst[-1]]
